# coding=utf8
# XBR Network - Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.

import binascii

import txaio

from crossbar._util import hlid

txaio.use_twisted()  # noqa
from txaio import make_logger, time_ns

import numpy as np

from autobahn.wamp import register, CallDetails
from autobahn.xbr import recover_eip712_market_member_login, is_cs_pubkey, is_signature, \
    is_address
from autobahn.util import generate_serial_number, without_0x
from autobahn.wamp.exception import ApplicationError

import cfxdb
from cfxdb.xbr import ActorType


class Authenticator:

    ERROR_INVALID_AUTH_REQUEST = 'xbr.marketmaker.error.invalid-auth-request'
    ERROR_INVALID_AUTH_REQUEST_MSG = 'invalid parameter(s) in authentication request: {}'

    log = make_logger()

    def __init__(self, db, market_session, reactor, market_oid):
        self._db = db
        self._schema = cfxdb.xbr.schema.Schema.attach(db)
        self._xbrmm = cfxdb.xbrmm.schema.Schema.attach(db)
        self._market_session = market_session
        self._market_oid = market_oid
        self._reactor = reactor
        self._pubkey_by_session = {}

    @register('xbr.marketmaker.authenticator.authenticate')
    async def _authenticate(self, realm, authid, details, call_details):
        self.log.info('{klass}.authenticate(realm="{realm}", authid="{authid}", details={details})',
                      klass=self.__class__.__name__,
                      realm=realm,
                      authid=authid,
                      details=details)

        if 'authmethod' not in details:
            msg = 'missing "authmethod" in authentication details (WAMP HELLO message details)'
            raise ApplicationError(self.ERROR_INVALID_AUTH_REQUEST, self.ERROR_INVALID_AUTH_REQUEST_MSG.format(msg))

        authmethod = details['authmethod']

        if authmethod != 'cryptosign':
            msg = 'authmethod "{}" not permissible'.format(authmethod)
            raise ApplicationError(self.ERROR_INVALID_AUTH_REQUEST, self.ERROR_INVALID_AUTH_REQUEST_MSG.format(msg))

        if 'authextra' not in details:
            msg = 'Must provide authextra for authmethod cryptosign'
            raise ApplicationError(self.ERROR_INVALID_AUTH_REQUEST, self.ERROR_INVALID_AUTH_REQUEST_MSG.format(msg))

        authextra = details['authextra']

        if 'pubkey' not in authextra:
            msg = 'missing public key in authextra for authmethod cryptosign'
            raise ApplicationError(self.ERROR_INVALID_AUTH_REQUEST, self.ERROR_INVALID_AUTH_REQUEST_MSG.format(msg))

        pubkey = authextra['pubkey']
        if isinstance(pubkey, str):
            pubkey = binascii.a2b_hex(without_0x(pubkey))
        assert is_cs_pubkey(pubkey)

        session_id = details['session']
        assert type(session_id) == int

        # FIXME: find a more elegant way to query the db.
        def get_actor(_txn, address):
            _actor = self._schema.actors[_txn, (self._market_oid, address, ActorType.PROVIDER)]
            if _actor:
                return _actor

            _actor = self._schema.actors[_txn, (self._market_oid, address, ActorType.CONSUMER)]
            if _actor:
                return _actor

            _actor = self._schema.actors[_txn, (self._market_oid, address, ActorType.PROVIDER_CONSUMER)]
            if _actor:
                return _actor

        if ('wallet_address' not in authextra or not authextra['wallet_address']) and \
                ('signature' not in authextra or not authextra['signature']):
            with self._db.begin() as txn:
                user_key = self._xbrmm.user_keys[txn, pubkey]
                actor = None
                if user_key:
                    actor = get_actor(txn, bytes(user_key.wallet_address))
                    if actor:
                        authrole = 'user'
                        authid = 'member-{}'.format(binascii.b2a_hex(user_key.wallet_address).decode())
                    else:
                        authrole = 'anonymous'
                        authid = 'anonymous-{}'.format(generate_serial_number())
                else:
                    authrole = 'anonymous'
                    authid = 'anonymous-{}'.format(generate_serial_number())

                self._pubkey_by_session[session_id] = pubkey

                auth = {
                    'pubkey': binascii.b2a_hex(pubkey),
                    'realm': realm,
                    'authid': authid,
                    'role': authrole,
                    'cache': True,
                    'extra': {
                        'actor_type': actor.actor_type if actor else 0
                    }
                }

                self.log.info('{klass}.authenticate(..) => {auth}', klass=self.__class__.__name__, auth=auth)

                return auth

        if ('wallet_address' not in authextra or not authextra['wallet_address']) or \
                ('signature' not in authextra or not authextra['signature']):
            msg = 'Should provide `pubkey`, `wallet_address` and `signature` in authextra ' \
                  'to authenticate new member. To authenticate existing member, only provide ' \
                  '`pubkey`'
            raise ApplicationError(self.ERROR_INVALID_AUTH_REQUEST, self.ERROR_INVALID_AUTH_REQUEST_MSG.format(msg))

        wallet_address = authextra['wallet_address']
        assert is_address(wallet_address)

        signature = authextra['signature']
        assert is_signature(signature)

        try:
            signer_address = recover_eip712_market_member_login(wallet_address, pubkey, signature)
        except Exception as e:
            self.log.warn('EIP712 signature recovery failed (wallet_adr={wallet_adr}): {err}',
                          wallet_adr=wallet_address,
                          err=str(e))
            raise ApplicationError('xbr.error.invalid_signature', 'EIP712 signature recovery failed ({})'.format(e))

        if signer_address != wallet_address:
            self.log.warn('EIP712 signature invalid: signer_address={signer_address}, wallet_adr={wallet_adr}',
                          signer_address=signer_address,
                          wallet_adr=wallet_address)
            raise ApplicationError('xbr.error.invalid_signature', 'EIP712 signature invalid')

        with self._db.begin(write=True) as txn:
            account = self._schema.members[txn, wallet_address]
            actor = None
            if account:
                actor = get_actor(txn, wallet_address)
                if actor:
                    user_key = self._xbrmm.user_keys[txn, pubkey]
                    if not user_key:
                        user_key = cfxdb.xbrmm.UserKey()
                        user_key.owner = account.account_oid
                        user_key.pubkey = pubkey
                        user_key.created = np.datetime64(txaio.time_ns(), 'ns')
                        user_key.wallet_address = wallet_address
                        user_key.signature = signature
                        self._xbrmm.user_keys[txn, pubkey] = user_key

                    self._pubkey_by_session[session_id] = pubkey

                    authrole = 'user'
                    # authid = 'member-{}'.format(account.account_oid)
                    # account.account_oid returns a pseudo value because
                    # the "emit" from the xbr contracts does not include
                    # account_oid in it, hence we don't really have that.
                    # To compensate that, we could include wallet address
                    # in authid, so that API calls could validate
                    # if the caller really is the "owner" of a resource.
                    authid = 'member-{}'.format(binascii.b2a_hex(wallet_address).decode())
                else:
                    authrole = 'anonymous'
                    authid = 'anonymous-{}'.format(generate_serial_number())

            else:
                authrole = 'anonymous'
                authid = 'anonymous-{}'.format(generate_serial_number())

        auth = {
            'pubkey': binascii.b2a_hex(pubkey),
            'realm': realm,
            'authid': authid,
            'role': authrole,
            'cache': True,
            'extra': {
                'actor_type': actor.actor_type if actor else 0
            }
        }

        self.log.info('{klass}.authenticate(..) => {auth}', klass=self.__class__.__name__, auth=auth)

        return auth

    @register('xbr.marketmaker.authenticator.logout')
    async def _logout(self, call_details: CallDetails):
        caller_session_id = call_details.caller

        caller_pubkey = self._pubkey_by_session.pop(caller_session_id, None)
        assert is_cs_pubkey(caller_pubkey)
        self.log.info('{klass}.logout_member with caller pubkey {caller_pubkey})',
                      klass=self.__class__.__name__,
                      caller_pubkey=hlid(binascii.b2a_hex(caller_pubkey).decode()))

        with self._db.begin(write=True) as txn:
            del self._xbrmm.user_keys[txn, caller_pubkey]

        logout_info = {
            'logged_out': time_ns(),
            'from_session': caller_session_id,
            'pubkey': caller_pubkey,
        }

        def kill():
            self._market_session.call('wamp.session.kill_by_authid', call_details.caller_authid)
            self.log.info('Ok, session {caller_session} logged out for client with pubkey {caller_pubkey} ',
                          caller_session=hlid(caller_session_id),
                          caller_pubkey=hlid(binascii.b2a_hex(caller_pubkey).decode()))

        # first return from this call, before killing its session ..
        self._reactor.callLater(0, kill)

        return logout_info



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import os
import shutil

from uuid import uuid4

from twisted.internet.endpoints import UNIXServerEndpoint, TCP4ServerEndpoint
from twisted.internet.endpoints import TCP4ClientEndpoint
from twisted.internet.selectreactor import SelectReactor
from twisted.internet.protocol import Factory
from twisted.protocols.wire import Echo
from twisted.python.runtime import platform

from crossbar.test import TestCase
from crossbar.common.twisted.endpoint import create_listening_endpoint_from_config
from crossbar.common.twisted.endpoint import create_connecting_endpoint_from_config

from txaio import make_logger


class ListeningEndpointTests(TestCase):

    log = make_logger()

    def setUp(self):
        self.cbdir = self.mktemp()
        os.makedirs(self.cbdir)
        return super(ListeningEndpointTests, self).setUp()

    def test_unix(self):
        """
        A config with type = "unix" will create an endpoint for a UNIX socket
        at the given path.
        """
        path = os.path.join("/", "tmp", uuid4().hex)
        self.addCleanup(os.remove, path)

        reactor = SelectReactor()
        config = {"type": "unix", "path": path}

        endpoint = create_listening_endpoint_from_config(config, self.cbdir, reactor, self.log)
        self.assertTrue(isinstance(endpoint, UNIXServerEndpoint))

        factory = Factory.forProtocol(Echo)
        endpoint.listen(factory)

        self.assertIn(factory, [getattr(x, "factory", None) for x in reactor.getReaders()])

    def test_unix_already_listening(self):
        """
        A config with type = "unix" will create an endpoint for a UNIX socket
        at the given path, and delete it if required.
        """
        path = os.path.join("/", "tmp", uuid4().hex)
        self.addCleanup(os.remove, path)

        # Something is already there
        open(path, "w").close()

        reactor = SelectReactor()
        config = {"type": "unix", "path": path}

        endpoint = create_listening_endpoint_from_config(config, self.cbdir, reactor, self.log)
        self.assertTrue(isinstance(endpoint, UNIXServerEndpoint))

        factory = Factory.forProtocol(Echo)
        endpoint.listen(factory)

        self.assertIn(factory, [getattr(x, "factory", None) for x in reactor.getReaders()])

    def test_unix_already_listening_cant_delete(self):
        """
        A config with type = "unix" will create an endpoint for a UNIX socket
        at the given path, and delete it if required. If it can't delete it, it
        will raise an exception.
        """
        parent_fp = os.path.join("/", "tmp", uuid4().hex)
        os.makedirs(parent_fp)
        fp = os.path.join(parent_fp, uuid4().hex)

        # Something is already there
        open(fp, "w").close()
        os.chmod(fp, 0o544)
        os.chmod(parent_fp, 0o544)

        reactor = SelectReactor()
        config = {"type": "unix", "path": fp}

        with self.assertRaises(OSError) as e:
            create_listening_endpoint_from_config(config, self.cbdir, reactor, self.log)
        self.assertEqual(e.exception.errno, 13)  # Permission Denied

        os.chmod(parent_fp, 0o700)
        shutil.rmtree(parent_fp)

    if platform.isWindows():
        _ = "Windows does not have UNIX sockets"
        test_unix.skip = _
        test_unix_already_listening.skip = _
        test_unix_already_listening_cant_delete.skip = _
        del _
    elif os.getuid() == 0:
        _ = "Cannot run as root"
        test_unix_already_listening_cant_delete.skip = _
        del _

    def test_twisted_client(self):
        reactor = SelectReactor()
        config = {
            "type": "twisted",
            "client_string": "tcp:host=127.0.0.1:port=9876",
        }

        endpoint = create_connecting_endpoint_from_config(config, self.cbdir, reactor, self.log)
        self.assertTrue(isinstance(endpoint, TCP4ClientEndpoint))

    def test_twisted_server(self):
        reactor = SelectReactor()
        config = {
            "type": "twisted",
            "server_string": "tcp:9876:interface=127.0.0.1",
        }

        endpoint = create_listening_endpoint_from_config(config, self.cbdir, reactor, self.log)
        self.assertTrue(isinstance(endpoint, TCP4ServerEndpoint))



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import json
import time
import os
from datetime import datetime
from selectors import DefaultSelector, EVENT_READ
from typing import List

import txaio

from twisted.internet.defer import inlineCallbacks
from twisted.internet import threads
from twisted.python.threadpool import ThreadPool

try:
    import docker
    HAS_DOCKER = True
except:
    HAS_DOCKER = False

if HAS_DOCKER:

    class Channel:
        log = txaio.make_logger()

        def __init__(self, selector, socket, id, tty_id, shell):
            self._selector = selector
            self._socket = socket
            self._id = id
            self._tty_id = tty_id
            self._shell = shell
            self.register()
            self.log.debug('Channel create, id={id}, tty_id={tty_id}', id=id, tty_id=tty_id)

        def set_tty(self, tty_id):
            self._tty_id = tty_id

        def register(self):
            self._key = self._selector.register(self._socket, events=EVENT_READ, data={'id': self._id})
            self.keepalive()

        def unregister(self):
            if self._key:
                self._selector.unregister(self._key.fileobj)
                self._key.fileobj.close()
                self._key = None

        def write(self, data):
            try:
                os.write(self._socket.fileno(), data.encode('utf-8'))
                self.keepalive()
                return
            except BrokenPipeError:
                self.log.debug('Someone stopped or reset the container [PIPE]!')
            except ValueError:
                self.log.debug('Someone stopped or reset the container [VALUE]!')
            self._shell = None

        def keepalive(self):
            self._lastseen = datetime.now()

        def expired(self, when):
            return (when - self._lastseen).seconds > 20

        def close(self):
            if self._shell:
                #
                #   Ok, hack alert !! Currently Docker exec doesn't support ANY
                #   mechanism for killing off exec'd commands .. so we're going to
                #   send a Control+C -> exit and hope for the best. If the user gets
                #   into trouble, they just have to restart the container.
                #
                self.write('\x03\n')
                self.write('exit\n')

            self.unregister()
            return self._key

    class Channels:
        log = txaio.make_logger()

        def __init__(self):
            self._channels = {}
            self._used_tty_ids = {}
            self._selector = DefaultSelector()
            self._last_channels = 0
            self._last_expired = 0
            self._next_tty_id = 0

        def get_next_id(self, id):
            if id and id in self._channels:
                return self.get_tty(id)
            tty_id = self._next_tty_id
            self._next_tty_id += 1
            return tty_id

        def get_tty(self, id):
            return self._channels[id]._tty_id if id in self._channels else -1

        def create(self, id, socket, tty_id, shell=False):
            self._used_tty_ids[tty_id] = True
            self._channels[id] = Channel(self._selector, socket, id, tty_id, shell)

        def exists(self, id):
            return id in self._channels

        def tty_exists(self, id):
            return id in self._used_tty_ids

        def set_tty(self, id, tty_id):
            self._channels[id].set_tty(tty_id)

        def silence(self, id):
            if id in self._channels:
                self._channels[id].unregister()

        def close(self, id):

            if id in self._channels:
                channel = self._channels[id]
                channel.close()
                if channel._tty_id in self._used_tty_ids:
                    del self._used_tty_ids[channel._tty_id]
                if id in self._channels:
                    del self._channels[id]
            else:
                self.log.error('attempt to close phantom channel')

        def write(self, id, data):
            self._channels[id].write(data)

        def select(self, timeout):
            return self._selector.select(timeout=timeout)

        def keepalive(self, id):
            self._channels[id].keepalive()

        def expire(self):
            channels = len(self._channels)
            expired = 0
            now = datetime.now()
            for id in dict(self._channels):
                if self._channels[id].expired(now):
                    expired += 1
                    self.close(id)
            if (channels != self._last_channels) or (expired != self._last_expired):
                self.log.info('Docker console expire - updated :: channels={}, expired={}'.format(channels, expired))
                self._last_channels = channels
                self._last_expired = expired

    class DockerClient:
        """
        Asynchronous Docker client (living on a background thread pool).
        """
        _docker = docker
        log = txaio.make_logger()

        CONSOLE_HISTORY = 60
        WAIT_TIMEOUT = 1
        EXCLUDE_DIRS_ANY = ['.cache']  # type: List[str]

        def __init__(self, reactor, controller):
            """
            Set up our async Docker interface.
            """
            self._reactor = reactor
            self._controller = controller
            self._finished = True
            self._channels = None
            self._threadpool = None
            self._events = None

        def console(self, section, status):
            self.log.info(f'docker - {section} - {status}')

        def startup(self):
            """
            Startup Docker client.
            """
            if not self._finished:
                self.log.warn('Docker client already running!')
                return

            self.console('module', 'starting')

            self._finished = False
            self._channels = Channels()

            # dedicated threadpool for docker work
            self._threadpool = ThreadPool(minthreads=4, maxthreads=100, name='docker')
            self._threadpool.start()

            # our 'events' pub/sub docker events emulator
            threads.deferToThreadPool(self._reactor, self._threadpool, self.events)

            # our 'logs' pub/sub docker console output
            threads.deferToThreadPool(self._reactor, self._threadpool, self.logs)

            # our 'keepalive' monitor
            threads.deferToThreadPool(self._reactor, self._threadpool, self.keepalive)

            self.console('module', 'started')

        def shutdown(self):
            """
            Shutdown Docker client.
            """
            if self._finished:
                self.console('module', 'already stopped')
                return

            self.console('module', 'stopping')
            self._finished = True
            self.console('keepalive', 'stopping')

            if self._events:
                self.console('events', 'stopping')
                self._events.close()

            if self._threadpool:
                self.console('threads', 'stopping')
                self._threadpool.stop()

            self.console('module', 'stopped')

        def keepalive(self):
            """
            Monitor all our active channels and expire any once keepalive's stop
            """
            self.console('keepalive', 'started')
            while not self._finished:
                self._channels.expire()
                for x in range(10):
                    if self._finished:
                        break
                    time.sleep(self.WAIT_TIMEOUT)
            self.console('keepalive', 'stopped')

        def logs(self):
            """
            Forward console logs from containers back into Crossbar
            """
            self.console('logs', 'started')
            while not self._finished:
                worklist = self._channels.select(self.WAIT_TIMEOUT)
                for (key, events) in worklist:
                    id = key.data['id']
                    line = os.read(key.fd, 8192)
                    if not line:
                        self._channels.silence(id)
                        continue
                    tty_id = self._channels.get_tty(id)
                    if tty_id >= 0:
                        self._reactor.callFromThread(
                            self._controller.publish,
                            f'crossbar.worker.{self._controller._uri_prefix}.docker.tty_{tty_id}',
                            {'line': line.decode('utf-8')})
                        time.sleep(0.1)
            self.console('logs', 'stopped')

        def events(self):
            """
            Called from node controller in a background thread to watch (blocking!)
            for Docker events and publish those as WAMP events on the main thread.
            """
            self.console('events', 'started')

            # DOCKER records logs with 1 second granularity, so it will potentiall have MANY lines with
            # the same timestamp. Asking for timestamp + delta is unsafe as it can only ask for the
            # next second, which will potentially lose all records logged which were effectively logged
            # against the previous second, but happened "after" that last call to "events". Docker "should"
            # log with time.time() or similar, but sadly ...

            # IF there is a problem with "events" in docker, it needs to be identified and fixed.
            # - this routine has been live on "demo1" for 9 months with no problems reported.

            while not self._finished:
                #
                #   "events" will close if docker is restarted, we aim to survive that event ...
                #
                try:
                    self._events = self._docker.from_env().events()
                    for event in self._events:
                        if self._finished:
                            break
                        event = json.loads(event, encoding='utf8')
                        ident = event.get('id')
                        if not ident:
                            continue
                        etype = event.get('Type')
                        eactn = event.get('Action')
                        topic = u'crossbar.worker.{}.docker.on_{}_{}'.format(self._controller._uri_prefix, etype,
                                                                             eactn)
                        if etype == 'container' and eactn == 'restart':
                            self.watch(ident, self._channels.get_tty(ident))
                        try:
                            payload = {'id': ident}
                            self.log.debug('publish : {topic} => {packet}', topic=topic, packet=payload)
                            if self._controller:
                                self._reactor.callFromThread(self._controller.publish, topic, payload)

                        except Exception as e:
                            self.log.error('Error: not able to handle event type :: {}'.format(topic))
                            print(e)

                except Exception as e:
                    self.log.error(f'error in "events" - {str(e)}')

            self.console('events', 'stopped')

        @inlineCallbacks
        def create(self, image, kwargs):
            """
            Create a new container and get it ready to run
            """
            def shim(image, **kwargs):
                client = self._docker.from_env()
                try:
                    container = client.containers.create(image, **kwargs)
                    return {'id': container.id}
                except docker.errors.ImageNotFound:
                    self.log.info('No Image ({image}) attempting to pull', image=image)
                    try:
                        client.images.pull(image)
                        container = client.containers.create(image, **kwargs)
                        return {'id': container.id}
                    except docker.errors.APIError:
                        raise Exception('Docker failed to pull ({image})', image=image)

            self.log.debug('docker create :: {image} -> {kw}', image=image, kw=kwargs)
            kwargs['detach'] = True
            kwargs['tty'] = True
            kwargs['stdin_open'] = True
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim, image, **kwargs))

        @inlineCallbacks
        def get_info(self):
            """
            Recover information about our docker installation.

            Shell command: ``crossbar shell --realm mrealm1 show docker node1``
            """
            def shim():
                return self._docker.from_env().info()

            self.log.debug('docker get_info')
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim))

        @inlineCallbacks
        def get_containers(self):
            """
            Recover a list of container ID's
            """
            def shim():
                return [c.id for c in self._docker.from_env().containers.list(all=True)]

            self.log.debug('docker get_containers')
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim))

        @inlineCallbacks
        def get_container(self, id):
            """
            Recover information about one specific container (by id)
            """
            def shim(id):
                try:
                    return self._docker.from_env().containers.get(id).attrs
                except Exception as e:
                    return {'error': 'unable to get container details', 'traceback': str(e)}

            self.log.debug('docker get_container -> {id}', id=id)
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim, id))

        @inlineCallbacks
        def get_images(self):
            """
            Recover a list of image ID's

            Shell command: ``crossbar shell --realm mrealm1 list docker-images node1``
            """
            def shim():
                return [c.id for c in self._docker.from_env().images.list()]

            self.log.debug('docker get_images')
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim))

        @inlineCallbacks
        def delete_image(self, id):
            """
            Purge old images
            """
            def shim():
                try:
                    return self._docker.from_env().images.remove(id)
                except Exception as e:
                    print(e)
                    print(dir(e))
                    return {'error': 'unable to remove image', 'traceback': str(e)}

            self.log.debug('docker delete_image')
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim))

        @inlineCallbacks
        def get_image(self, id):
            """
            Recover information about one specific image (by id)

            Shell command: ``crossbar shell --realm mrealm1 show docker-image node1 4bbb66``
            """
            def shim(id):
                try:
                    return self._docker.from_env().images.get(id).attrs
                except Exception as e:
                    return {'error': 'unable to get image', 'traceback': str(e)}

            self.log.debug('docker get_image -> {id}', id=id)
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim, id))

        @inlineCallbacks
        def df(self):
            """
            Get information relating to docker's usage of available storage
            """
            def shim():
                return self._docker.from_env().df()

            self.log.debug('docker df')
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim))

        @inlineCallbacks
        def ping(self):
            """
            Bounce a message off docket to see if it's running
            """
            def shim():
                try:
                    return self._docker.from_env().ping()
                except Exception:
                    return False

            self.log.debug('docker ping')
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim))

        @inlineCallbacks
        def version(self):
            """
            Get the version information of our docker instance
            """
            def shim():
                return self._docker.from_env().version()

            self.log.debug('docker version')
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim))

        @inlineCallbacks
        def container(self, id, cmd):
            """
            Operate on a specific container (by id)
            """
            def shim(id, cmd):
                container = self._docker.from_env().containers.get(id)
                if hasattr(container, cmd):
                    return getattr(container, cmd)()
                raise Exception('no such command :: {}'.format(cmd))

            self.log.debug('docker container -> {id} + {cmd}', id=id, cmd=cmd)
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim, id, cmd))

        @inlineCallbacks
        def start(self, id):
            """
            Specific container routine that needs a non-default timeout
            """
            def shim(id):
                container = self._docker.from_env().containers.get(id)
                status = container.start()
                tty_id = self._channels.get_tty(id)
                if tty_id >= 0:
                    client = docker.APIClient()
                    params = {'stdin': 1, 'stdout': 1, 'stderr': 1, 'stream': 1, 'timestamps': 0, 'logs': 0}
                    socket = client.attach_socket(id, params)
                    self._channels.close(id)
                    self._channels.create(id, socket, tty_id)
                    status = {'status': 'OK', 'id': id, 'tty_id': tty_id}
                return status

            self.log.debug('docker container start -> {id}', id=id)
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim, id))

        @inlineCallbacks
        def restart(self, id):
            """
            Specific container routine that needs a non-default timeout
            """
            def shim(id):
                container = self._docker.from_env().containers.get(id)
                try:
                    container.restart(timeout=1)
                except Exception as e:
                    self.log.error('Exception while trying to restart container')
                    self.log.error(str(e))
                tty_id = self._channels.get_tty(id)
                client = self._docker.APIClient()
                params = {'stdin': 1, 'stdout': 1, 'stderr': 1, 'stream': 1, 'timestamps': 0, 'logs': 0}
                socket = client.attach_socket(id, params)
                self._channels.close(id)
                self._channels.create(id, socket, tty_id)
                return {'status': 'OK', 'id': id, 'tty_id': tty_id}

            self.log.info('docker container restart -> {id}', id=id)
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim, id))

        @inlineCallbacks
        def image(self, id, cmd):
            """
            Operate on a specific image (by id)
            """
            def shim(id, cmd):
                image = self._docker.from_env().images.get(id)
                if hasattr(image, cmd):
                    return getattr(image, cmd)()
                raise Exception('no such command :: {}'.format(cmd))

            self.log.debug('docker image -> {id} + {cmd}', id=id, cmd=cmd)
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim, id, cmd))

        @inlineCallbacks
        def prune(self, filter):
            """
            Prune docker images
            """
            def shim(filter):
                return self._docker.from_env().images.prune(filter)

            self.log.debug('docker prune -> {filter}', filter=filter)
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim, filter))

        @inlineCallbacks
        def backlog(self, id):
            """
            Request historical console logs / buffer
            """
            def shim(id):
                if not self._channels.exists(id):
                    return {'status': 'NOTFOUND'}

                client = self._docker.from_env()
                try:
                    container = client.containers.get(id)
                # FIXME: NotFound
                except:
                    return {'status': 'OK', 'packet': ''}

                lines = container.logs(stdout=1, stderr=1, stream=0, timestamps=1, tail=60)
                return {'status': 'OK', 'packet': lines[-16384:].decode('utf-8')}

            self.log.debug('docker backlog -> {id}', id=id)
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim, id))

        def request_tty(self, id):
            return {'status': 'OK', 'tty_id': self._channels.get_next_id(id)}

        @inlineCallbacks
        def watch(self, id, tty_id):
            """
            Watch the console of the specified container
            """
            def shim(id):
                try:
                    client = self._docker.from_env()
                    container = client.containers.get(id)
                except docker.errors.NotFound:
                    return {'status': 'NOTFOUND', 'packet': ''}

                if self._channels.exists(id) or self._channels.tty_exists(tty_id):
                    self._channels.set_tty(id, tty_id)
                    buffer = container.logs(stdout=1, stderr=1, stream=0, timestamps=0, tail=self.CONSOLE_HISTORY)
                else:
                    client = self._docker.APIClient()
                    params = {'stdin': 1, 'stdout': 1, 'stderr': 1, 'stream': 1, 'timestamps': 0, 'logs': 0}
                    socket = client.attach_socket(id, params)
                    buffer = container.logs(stdout=1, stderr=1, stream=0, timestamps=0, tail=self.CONSOLE_HISTORY)
                    self._channels.create(id, socket, tty_id)

                buffer = buffer.decode('utf-8')
                # attempt to clean broken ESC sequence
                if len(buffer) > 16384:
                    buffer = buffer[-16384:]
                    for i in range(12):
                        if ord(buffer[i]) == 27:
                            buffer = buffer[i:]
                            break
                return {'status': 'OK', 'id': id, 'tty_id': tty_id, 'buffer': buffer}

            self.log.debug('docker watch -> {id}', id=id)
            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim, id))

        def keystroke(self, id, data):
            """
            Enter a new keystroke into a container console
            """
            if not self._channels.exists(id):
                return {'status': 'NOTFOUND'}

            if isinstance(data, list):
                for item in data:
                    if item['action'] == 'keepalive':
                        self._channels.keepalive(id)
                    elif item['action'] == 'size_console':
                        client = self._docker.from_env()
                        container = client.containers.get(id)
                        container.resize(item['rows'], item['cols'])
                    elif item['action'] == 'size_shell':
                        client = self._docker.APIClient()
                        client.exec_resize(id, item['rows'], item['cols'])
                    elif item['action'] == 'close':
                        self._channels.close(id)
                    else:
                        self.log.error('unknown keystroke command: {cmd}', cmd=item)
                return

            self._channels.write(id, data)
            return {'status': 'OK'}

        @inlineCallbacks
        def shell(self, container, tty_id, kwargs={}):
            """
            Execute a shell in a running container
            """
            client = self._docker.APIClient()

            def shim(image, **kwargs):
                kwargs['tty'] = True
                kwargs['stdin'] = True
                cmd = '/bin/bash'
                return client.exec_create(container, cmd, **kwargs)

            self.log.debug('docker shell :: {container} -> {kw}', container=container, kw=kwargs)
            execId = (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim, container, **kwargs))
            id = execId.get('Id')

            def shim2():
                socket = client.exec_start(id, detach=False, tty=True, socket=True)
                self._channels.create(id, socket, tty_id, True)
                return {'status': 'OK', 'id': id}

            return (yield threads.deferToThreadPool(self._reactor, self._threadpool, shim2))

        def fs_root(self, id, path):
            while path and path[0] == '/':
                path = path[1:]
            container = self._docker.from_env().containers.get(id)
            for point in container.attrs.get('Mounts', []):
                dst = point.get('Destination', '')
                src = point.get('Source', '')
                if f'/{path}'.startswith(f'{dst}/'):
                    path = "/".join(path.split('/')[1:])
                    return os.path.join(src, path)
            raise Exception(f'invalid path "{path}"')

        def fs_open(self, id, path):
            """
            Read the filesystem structure for the given container
            """
            while path and path[0] == '/':
                path = path[1:]
            files = []
            dirs = []
            container = self._docker.from_env().containers.get(id)
            if not path:
                for point in container.attrs.get('Mounts', []):
                    dirs.append(point.get('Destination'))
            else:
                if path[:-1] != '/':
                    path += '/'
                for point in container.attrs.get('Mounts', []):
                    dst = point.get('Destination', '')
                    src = point.get('Source', '')
                    print(f'path={path} dst={dst}')
                    if not f'/{path}'.startswith(f'{dst}/'):
                        continue
                    path = "/".join(path.split('/')[1:])
                    root = os.path.join(src, path)
                    print(f'Root={root}')
                    with os.scandir(root) as iterator:
                        for entry in iterator:
                            if entry.is_file():
                                files.append(entry.name)
                            elif entry.is_dir():
                                if entry.name not in self.EXCLUDE_DIRS_ANY:
                                    dirs.append(entry.name)
            dirs.sort()
            files.sort()
            return {'dirs': dirs, 'files': files}

        def fs_get(self, id, path):
            """
            Recover a file from a container filesystem
            """
            with open(self.fs_root(id, path)) as io:
                return {'data': io.read()}

        def fs_put(self, id, path, data):
            """
            Store a file into a Docker container
            """
            with open(self.fs_root(id, path), 'w') as io:
                io.write(data)



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import json
from collections.abc import Sequence

import crossbar
from crossbar.test import TestCase
from crossbar.common import checkconfig

_DEFAULT_PERSONALITY_CLASS = crossbar.personalities()['standalone']


class CheckDictArgsTests(TestCase):
    """
    Tests for L{crossbar.common.checkconfig.check_dict_args}.
    """
    def test_sequence_string(self):
        """
        A Sequence should not imply we accept strings
        """
        with self.assertRaises(checkconfig.InvalidConfigException) as e:
            checkconfig.check_dict_args({"foo": (True, [Sequence])}, {"foo": "not really a Sequence"},
                                        "Nice message for the user")
        self.assertEqual(
            "Nice message for the user - invalid type str encountered for "
            "attribute 'foo', must be one of (Sequence)",
            str(e.exception),
        )

    def test_sequence_list(self):
        """
        A Sequence should accept list
        """
        checkconfig.check_dict_args({"foo": (True, [Sequence])}, {"foo": ["a", "real", "sequence"]},
                                    "Nice message for the user")
        # should work, with no exceptions

    def test_notDict(self):
        """
        A non-dict passed in as the config will raise a
        L{checkconfig.InvalidConfigException}.
        """
        with self.assertRaises(checkconfig.InvalidConfigException) as e:
            checkconfig.check_dict_args({}, [], "msghere")

        self.assertEqual("msghere - invalid type for configuration item - expected dict, got list", str(e.exception))

    def test_wrongType(self):
        """
        The wrong type (as defined in the spec) passed in the config will raise
        a L{checkconfig.InvalidConfigException}.
        """
        with self.assertRaises(checkconfig.InvalidConfigException) as e:
            checkconfig.check_dict_args({"foo": (False, [list, set])}, {"foo": {}}, "msghere")

        self.assertEqual(("msghere - invalid type dict encountered for "
                          "attribute 'foo', must be one of (list, set)"), str(e.exception))


class CheckContainerTests(TestCase):
    """
    Tests for L{crossbar.common.checkconfig.check_container}.
    """
    def setUp(self):
        self.personality = _DEFAULT_PERSONALITY_CLASS
        return super(TestCase, self).setUp()

    def test_validTemplate_hello(self):
        """
        The config provided by the hello:python template should validate
        successfully.
        """
        config = json.loads('''{
            "type": "container",
            "options": {
                "pythonpath": [".."]
            },
            "components": [
                {
                    "type": "class",
                    "classname": "hello.hello.AppSession",
                    "realm": "realm1",
                    "transport": {
                        "type": "websocket",
                        "endpoint": {
                            "type": "tcp",
                            "host": "127.0.0.1",
                            "port": 8080
                        },
                        "url": "ws://127.0.0.1:8080/ws"
                    }
                }
            ]
        }''')
        self.personality.check_container(self.personality, config)

    def test_extraKeys(self):
        """
        A component with extra keys will fail.
        """
        config = json.loads('''{
            "type": "container",
            "options": {
                "pythonpath": [".."]
            },
            "components": [
                {
                    "type": "class",
                    "classname": "hello.hello.AppSession",
                    "realm": "realm1",
                    "woooooo": "bar",
                    "transport": {
                        "type": "websocket",
                        "endpoint": {
                            "type": "tcp",
                            "host": "127.0.0.1",
                            "port": 8080
                        },
                        "url": "ws://127.0.0.1:8080/ws"
                    }
                }
            ]
        }''')
        with self.assertRaises(checkconfig.InvalidConfigException) as e:
            self.personality.check_container(self.personality, config)

        self.assertIn("encountered unknown attribute 'woooooo'", str(e.exception))

    def test_requiredKeys(self):
        """
        A component with missing keys fails.
        """
        config = json.loads('''{
            "type": "container",
            "options": {
                "pythonpath": [".."]
            },
            "components": [
                {
                    "type": "class",
                    "classname": "hello.hello.AppSession",
                    "realm": "realm1"
                }
            ]
        }''')
        with self.assertRaises(checkconfig.InvalidConfigException) as e:
            self.personality.check_container(self.personality, config)

        self.assertIn("invalid component configuration - missing mandatory attribute 'transport'", str(e.exception))


class CheckEndpointTests(TestCase):
    """
    check_listening_endpoint and check_connecting_endpoint
    """
    def setUp(self):
        self.personality = _DEFAULT_PERSONALITY_CLASS
        return super(TestCase, self).setUp()

    def test_twisted_client_error(self):
        config = {
            "type": "twisted",
            "client_string": 1000,
        }

        with self.assertRaises(checkconfig.InvalidConfigException) as ctx:
            self.personality.check_connecting_endpoint(self.personality, config)
        self.assertTrue("in Twisted endpoint must be str" in str(ctx.exception))

    def test_twisted_server_error(self):
        config = {
            "type": "twisted",
            "server_string": 1000,
        }

        with self.assertRaises(checkconfig.InvalidConfigException) as ctx:
            self.personality.check_listening_endpoint(self.personality, config)
        self.assertTrue("in Twisted endpoint must be str" in str(ctx.exception))

    def test_twisted_server_missing_arg(self):
        config = {"type": "twisted"}

        with self.assertRaises(checkconfig.InvalidConfigException) as ctx:
            self.personality.check_listening_endpoint(self.personality, config)
        self.assertTrue("mandatory attribute 'server_string'" in str(ctx.exception))

    def test_twisted_client_missing_arg(self):
        config = {"type": "twisted"}

        with self.assertRaises(checkconfig.InvalidConfigException) as ctx:
            self.personality.check_connecting_endpoint(self.personality, config)
        self.assertTrue("mandatory attribute 'client_string'" in str(ctx.exception))


class CheckWebsocketTests(TestCase):
    def setUp(self):
        self.personality = _DEFAULT_PERSONALITY_CLASS
        return super(TestCase, self).setUp()

    def test_tiny_timeout_auto_ping(self):
        options = dict(auto_ping_timeout=12)

        with self.assertRaises(checkconfig.InvalidConfigException) as ctx:
            checkconfig.check_websocket_options(options)

        self.assertTrue("'auto_ping_timeout' is in milliseconds" in str(ctx.exception))


class CheckRealmTests(TestCase):
    """
    Tests for check_router_realm, check_router_realm_role
    """
    def setUp(self):
        self.personality = _DEFAULT_PERSONALITY_CLASS
        return super(TestCase, self).setUp()

    def test_dynamic_authorizer(self):
        config_realm = {"name": "realm1", "roles": [{"name": "dynamic", "authorizer": "com.example.foo"}]}

        self.personality.check_router_realm(self.personality, config_realm)

    def test_static_permissions(self):
        config_realm = {
            "name":
            "realm1",
            "roles": [{
                "name":
                "backend",
                "permissions": [{
                    "uri": "*",
                    "allow": {
                        "publish": True,
                        "subscribe": True,
                        "call": True,
                        "register": True
                    }
                }]
            }]
        }

        self.personality.check_router_realm(self.personality, config_realm)

    def test_static_permissions_invalid_uri(self):
        config_realm = {
            "name":
            "realm1",
            "roles": [{
                "name":
                "backend",
                "permissions": [{
                    "uri": "*foo",
                    "allow": {
                        "publish": True,
                        "subscribe": True,
                        "call": True,
                        "register": True
                    }
                }]
            }]
        }

        self.assertRaises(
            checkconfig.InvalidConfigException,
            self.personality.check_router_realm,
            self.personality,
            config_realm,
        )

    def test_static_permissions_and_authorizer(self):
        config_realm = {
            "name": "realm1",
            "roles": [{
                "name": "backend",
                "authorizer": "com.example.foo",
                "permissions": [],
            }]
        }

        self.assertRaises(
            checkconfig.InvalidConfigException,
            self.personality.check_router_realm,
            self.personality,
            config_realm,
        )

    def test_static_permissions_isnt_list(self):
        config_realm = {
            "name": "realm1",
            "roles": [{
                "name": "backend",
                "permissions": {},
            }]
        }

        self.assertRaises(
            checkconfig.InvalidConfigException,
            self.personality.check_router_realm,
            self.personality,
            config_realm,
        )

    def test_static_permissions_not_dict(self):
        config_realm = {"name": "realm1", "roles": [{"name": "backend", "permissions": ["not a dict"]}]}

        self.assertRaises(
            checkconfig.InvalidConfigException,
            self.personality.check_router_realm,
            self.personality,
            config_realm,
        )

    def test_static_permissions_lacks_uri(self):
        config_realm = {
            "name":
            "realm1",
            "roles": [{
                "name": "backend",
                "permissions": [{
                    "allow": {
                        "publish": True,
                        "subscribe": True,
                        "call": True,
                        "register": True
                    }
                }]
            }]
        }

        self.assertRaises(
            checkconfig.InvalidConfigException,
            self.personality.check_router_realm,
            self.personality,
            config_realm,
        )

    def test_static_permissions_uri_not_a_string(self):
        config_realm = {"name": "realm1", "roles": [{"name": "backend", "permissions": [{"uri": {}}]}]}

        self.assertRaises(
            checkconfig.InvalidConfigException,
            self.personality.check_router_realm,
            self.personality,
            config_realm,
        )


class CheckOnion(TestCase):
    def setUp(self):
        self.personality = _DEFAULT_PERSONALITY_CLASS
        return super(TestCase, self).setUp()

    def test_unknown_attr(self):
        with self.assertRaises(checkconfig.InvalidConfigException) as ctx:
            self.personality.check_listening_endpoint_onion(self.personality, {
                "type": "onion",
                "foo": 42,
            })
        self.assertIn("unknown attribute", str(ctx.exception))

    def test_success(self):
        self.personality.check_listening_endpoint_onion(
            self.personality, {
                "type": "onion",
                "private_key_file": "something",
                "port": 1234,
                "tor_control_endpoint": {
                    "type": "unix",
                    "path": "/dev/null",
                }
            })

    def test_port_wrong_type(self):
        with self.assertRaises(checkconfig.InvalidConfigException) as ctx:
            self.personality.check_listening_endpoint_onion(self.personality, {
                "type": "onion",
                "port": "1234",
            })
        self.assertIn("invalid type", str(ctx.exception))
        self.assertIn("encountered for attribute 'port'", str(ctx.exception))



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import json
import os
import sys

from twisted.internet.selectreactor import SelectReactor
from twisted.internet.task import LoopingCall

from crossbar.node import main
from .test_cli import CLITestBase

# Turn this to `True` to print the stdout/stderr of the Crossbars spawned
DEBUG = False


def make_lc(self, reactor, func):

    if DEBUG:
        self.stdout_length = 0
        self.stderr_length = 0

    def _(lc, reactor):
        if DEBUG:
            stdout = self.stdout.getvalue()
            stderr = self.stderr.getvalue()

            if self.stdout.getvalue()[self.stdout_length:]:
                print(self.stdout.getvalue()[self.stdout_length:], file=sys.__stdout__)
            if self.stderr.getvalue()[self.stderr_length:]:
                print(self.stderr.getvalue()[self.stderr_length:], file=sys.__stderr__)

            self.stdout_length = len(stdout)
            self.stderr_length = len(stderr)

        return func(lc, reactor)

    lc = LoopingCall(_)
    lc.a = (lc, reactor)
    lc.clock = reactor
    lc.start(0.1)
    return lc


class ContainerRunningTests(CLITestBase):
    def setUp(self):

        CLITestBase.setUp(self)

        # Set up the configuration directories
        self.cbdir = os.path.abspath(self.mktemp())
        os.mkdir(self.cbdir)
        self.config = os.path.abspath(os.path.join(self.cbdir, "config.json"))
        self.code_location = os.path.abspath(self.mktemp())
        os.mkdir(self.code_location)

    def _start_run(self, config, app, stdout_expected, stderr_expected, end_on):

        with open(self.config, "wb") as f:
            f.write(json.dumps(config, ensure_ascii=False).encode('utf8'))

        with open(self.code_location + "/myapp.py", "w") as f:
            f.write(app)

        reactor = SelectReactor()

        make_lc(self, reactor, end_on)

        # In case it hard-locks
        reactor.callLater(self._subprocess_timeout, reactor.stop)

        main.main("crossbar", ["start", "--cbdir={}".format(self.cbdir), "--logformat=syslogd"], reactor=reactor)

        out = self.stdout.getvalue()
        err = self.stderr.getvalue()
        for i in stdout_expected:
            if i not in out:
                self.fail("Error: '{}' not in:\n{}".format(i, out))

        for i in stderr_expected:
            if i not in err:
                self.fail("Error: '{}' not in:\n{}".format(i, err))

    def test_start_run(self):
        """
        A basic start, that enters the reactor.
        """
        expected_stdout = ["Entering reactor event loop", "Loaded the component!"]
        expected_stderr = []

        def _check(lc, reactor):
            if "Loaded the component!" in self.stdout.getvalue():
                lc.stop()
                try:
                    reactor.stop()
                except:
                    pass

        config = {
            "controller": {},
            "workers": [{
                "type":
                "router",
                "options": {
                    "pythonpath": ["."]
                },
                "realms": [{
                    "name":
                    "realm1",
                    "roles": [{
                        "name":
                        "anonymous",
                        "permissions": [{
                            "uri": "*",
                            "publish": True,
                            "subscribe": True,
                            "call": True,
                            "register": True
                        }]
                    }]
                }],
                "transports": [{
                    "type": "web",
                    "endpoint": {
                        "type": "tcp",
                        "port": 8080
                    },
                    "paths": {
                        "/": {
                            "directory": ".",
                            "type": "static"
                        },
                        "ws": {
                            "type": "websocket"
                        }
                    }
                }]
            }, {
                "type":
                "container",
                "options": {
                    "pythonpath": [self.code_location]
                },
                "components": [{
                    "type": "class",
                    "classname": "myapp.MySession",
                    "realm": "realm1",
                    "transport": {
                        "type": "websocket",
                        "endpoint": {
                            "type": "tcp",
                            "host": "127.0.0.1",
                            "port": 8080
                        },
                        "url": "ws://127.0.0.1:8080/ws"
                    }
                }]
            }]
        }

        myapp = """#!/usr/bin/env python
from twisted.logger import Logger
from autobahn.twisted.wamp import ApplicationSession
from autobahn.wamp.exception import ApplicationError

class MySession(ApplicationSession):

    log = Logger()

    def onJoin(self, details):
        self.log.info("Loaded the component!")
"""

        self._start_run(config, myapp, expected_stdout, expected_stderr, _check)

    def test_start_run_guest(self):
        """
        A basic start of a guest.
        """
        expected_stdout = ["Entering reactor event loop", "Loaded the component!"]
        expected_stderr = []

        def _check(lc, reactor):
            if "Loaded the component!" in self.stdout.getvalue():
                lc.stop()
                try:
                    reactor.stop()
                except:
                    pass

        config = {
            "controller": {},
            "workers": [{
                "type":
                "router",
                "options": {
                    "pythonpath": ["."]
                },
                "realms": [{
                    "name":
                    "realm1",
                    "roles": [{
                        "name":
                        "anonymous",
                        "permissions": [{
                            "uri": "*",
                            "publish": True,
                            "subscribe": True,
                            "call": True,
                            "register": True
                        }]
                    }]
                }],
                "transports": [{
                    "type": "web",
                    "endpoint": {
                        "type": "tcp",
                        "port": 8080
                    },
                    "paths": {
                        "/": {
                            "directory": ".",
                            "type": "static"
                        },
                        "ws": {
                            "type": "websocket"
                        }
                    }
                }]
            }, {
                "type": "guest",
                "executable": sys.executable,
                "arguments": [os.path.join(self.code_location, "myapp.py")]
            }]
        }

        myapp = """#!/usr/bin/env python
print("Loaded the component!")
"""

        self._start_run(config, myapp, expected_stdout, expected_stderr, _check)

    def test_start_utf8_logging(self):
        """
        Logging things that are UTF8 but not Unicode should work fine.
        """
        expected_stdout = ["Entering reactor event loop", "\u2603"]
        expected_stderr = []

        def _check(lc, reactor):
            if "\u2603" in self.stdout.getvalue():
                lc.stop()
                try:
                    reactor.stop()
                except:
                    pass

        config = {
            "controller": {},
            "workers": [{
                "type":
                "router",
                "options": {
                    "pythonpath": ["."]
                },
                "realms": [{
                    "name":
                    "realm1",
                    "roles": [{
                        "name":
                        "anonymous",
                        "permissions": [{
                            "uri": "*",
                            "publish": True,
                            "subscribe": True,
                            "call": True,
                            "register": True
                        }]
                    }]
                }],
                "transports": [{
                    "type": "web",
                    "endpoint": {
                        "type": "tcp",
                        "port": 8080
                    },
                    "paths": {
                        "/": {
                            "directory": ".",
                            "type": "static"
                        },
                        "ws": {
                            "type": "websocket"
                        }
                    }
                }]
            }, {
                "type":
                "container",
                "options": {
                    "pythonpath": [self.code_location]
                },
                "components": [{
                    "type": "class",
                    "classname": "myapp.MySession",
                    "realm": "realm1",
                    "transport": {
                        "type": "websocket",
                        "endpoint": {
                            "type": "tcp",
                            "host": "127.0.0.1",
                            "port": 8080
                        },
                        "url": "ws://127.0.0.1:8080/ws"
                    }
                }]
            }]
        }

        myapp = """#!/usr/bin/env python
from twisted.logger import Logger
from autobahn.twisted.wamp import ApplicationSession
from autobahn.wamp.exception import ApplicationError

class MySession(ApplicationSession):

    log = Logger()

    def onJoin(self, details):
        self.log.info("\\u2603")
"""

        self._start_run(config, myapp, expected_stdout, expected_stderr, _check)

    def test_run_exception_utf8(self):
        """
        Raising an ApplicationError with Unicode will raise that error through
        to the caller.
        """
        config = {
            "workers": [{
                "type":
                "router",
                "realms": [{
                    "name":
                    "realm1",
                    "roles": [{
                        "name":
                        "anonymous",
                        "permissions": [{
                            "uri": "*",
                            "publish": True,
                            "subscribe": True,
                            "call": True,
                            "register": True
                        }]
                    }]
                }],
                "transports": [{
                    "type": "web",
                    "endpoint": {
                        "type": "tcp",
                        "port": 8080
                    },
                    "paths": {
                        "/": {
                            "type": "static",
                            "directory": ".."
                        },
                        "ws": {
                            "type": "websocket"
                        }
                    }
                }]
            }, {
                "type":
                "container",
                "options": {
                    "pythonpath": [self.code_location]
                },
                "components": [{
                    "type": "class",
                    "classname": "myapp.MySession",
                    "realm": "realm1",
                    "transport": {
                        "type": "websocket",
                        "endpoint": {
                            "type": "tcp",
                            "host": "127.0.0.1",
                            "port": 8080
                        },
                        "url": "ws://127.0.0.1:8080/ws"
                    }
                }]
            }]
        }

        myapp = """from twisted.logger import Logger
from autobahn.twisted.wamp import ApplicationSession
from autobahn.wamp.exception import ApplicationError
from twisted.internet.defer import inlineCallbacks

class MySession(ApplicationSession):

    log = Logger()

    @inlineCallbacks
    def onJoin(self, details):

        def _err():
            raise ApplicationError("com.example.error.form_error", "\\u2603")
        e = yield self.register(_err, 'com.example.err')

        try:
            yield self.call('com.example.err')
        except ApplicationError as e:
            assert e.args[0] == "\\u2603"
            print("Caught error:", e)
        except:
            print('other err:', e)

        self.log.info("Loaded the component")
"""

        expected_stdout = ["Loaded the component", "\u2603", "Caught error:"]
        expected_stderr = []

        def _check(lc, reactor):
            if "Loaded the component" in self.stdout.getvalue():
                lc.stop()
                try:
                    reactor.stop()
                except:
                    pass

        self._start_run(config, myapp, expected_stdout, expected_stderr, _check)

    def test_failure1(self):

        config = {
            "workers": [{
                "type":
                "router",
                "realms": [{
                    "name":
                    "realm1",
                    "roles": [{
                        "name":
                        "anonymous",
                        "permissions": [{
                            "uri": "*",
                            "publish": True,
                            "subscribe": True,
                            "call": True,
                            "register": True
                        }]
                    }]
                }],
                "transports": [{
                    "type": "web",
                    "endpoint": {
                        "type": "tcp",
                        "port": 8080
                    },
                    "paths": {
                        "/": {
                            "type": "static",
                            "directory": ".."
                        },
                        "ws": {
                            "type": "websocket"
                        }
                    }
                }]
            }, {
                "type":
                "container",
                "components": [{
                    "type": "class",
                    "classname": "myapp.MySession",
                    "realm": "realm1",
                    "transport": {
                        "type": "websocket",
                        "endpoint": {
                            "type": "tcp",
                            "host": "127.0.0.1",
                            "port": 8080
                        },
                        "url": "ws://127.0.0.1:8080/ws"
                    }
                }]
            }]
        }

        myapp = """from twisted.logger import Logger
from autobahn.twisted.wamp import ApplicationSession

class MySession(ApplicationSession):

    log = Logger()

    def __init__(self, config):
        self.log.info("MySession.__init__()")
        ApplicationSession.__init__(self, config)

    def onJoin(self, details):
        self.log.info("MySession.onJoin()")
"""

        expected_stdout = []
        expected_stderr = ["No module named"]

        def _check(_1, _2):
            pass

        self._start_run(config, myapp, expected_stdout, expected_stderr, _check)

    def test_failure2(self):

        config = {
            "workers": [{
                "type":
                "router",
                "realms": [{
                    "name":
                    "realm1",
                    "roles": [{
                        "name":
                        "anonymous",
                        "permissions": [{
                            "uri": "*",
                            "publish": True,
                            "subscribe": True,
                            "call": True,
                            "register": True
                        }]
                    }]
                }],
                "transports": [{
                    "type": "web",
                    "endpoint": {
                        "type": "tcp",
                        "port": 8080
                    },
                    "paths": {
                        "/": {
                            "type": "static",
                            "directory": ".."
                        },
                        "ws": {
                            "type": "websocket"
                        }
                    }
                }]
            }, {
                "type":
                "container",
                "options": {
                    "pythonpath": [self.code_location]
                },
                "components": [{
                    "type": "class",
                    "classname": "myapp.MySession2",
                    "realm": "realm1",
                    "transport": {
                        "type": "websocket",
                        "endpoint": {
                            "type": "tcp",
                            "host": "127.0.0.1",
                            "port": 8080
                        },
                        "url": "ws://127.0.0.1:8080/ws"
                    }
                }]
            }]
        }

        myapp = """
from twisted.logger import Logger
from autobahn.twisted.wamp import ApplicationSession

class MySession(ApplicationSession):

    log = Logger()

    def __init__(self, config):
        self.log.info("MySession.__init__()")
        ApplicationSession.__init__(self, config)

    def onJoin(self, details):
        self.log.info("MySession.onJoin()")
"""

        def _check(_1, _2):
            pass

        expected_stdout = []
        if sys.version_info >= (3, 5):
            expected_stderr = ["module 'myapp' has no attribute 'MySession2'"]
        else:
            expected_stderr = ["'module' object has no attribute 'MySession2'"]

        self._start_run(config, myapp, expected_stdout, expected_stderr, _check)

    def test_failure3(self):

        config = {
            "workers": [{
                "type":
                "router",
                "realms": [{
                    "name":
                    "realm1",
                    "roles": [{
                        "name":
                        "anonymous",
                        "permissions": [{
                            "uri": "*",
                            "publish": True,
                            "subscribe": True,
                            "call": True,
                            "register": True
                        }]
                    }]
                }],
                "transports": [{
                    "type": "web",
                    "endpoint": {
                        "type": "tcp",
                        "port": 8080
                    },
                    "paths": {
                        "/": {
                            "type": "static",
                            "directory": ".."
                        },
                        "ws": {
                            "type": "websocket"
                        }
                    }
                }]
            }, {
                "type":
                "container",
                "options": {
                    "pythonpath": [self.code_location]
                },
                "components": [{
                    "type": "class",
                    "classname": "myapp.MySession",
                    "realm": "realm1",
                    "transport": {
                        "type": "websocket",
                        "endpoint": {
                            "type": "tcp",
                            "host": "127.0.0.1",
                            "port": 8080
                        },
                        "url": "ws://127.0.0.1:8080/ws"
                    }
                }]
            }]
        }

        myapp = """
from twisted.logger import Logger
from autobahn.twisted.wamp import ApplicationSession

class MySession(ApplicationSession):

    log = Logger()

    def __init__(self, config):
        a = 1 / 0
        self.log.info("MySession.__init__()")
        ApplicationSession.__init__(self, config)

    def onJoin(self, details):
        self.log.info("MySession.onJoin()")
"""

        def _check(_1, _2):
            pass

        expected_stdout = []
        expected_stderr = ["Component instantiation failed", "division by zero"]

        self._start_run(config, myapp, expected_stdout, expected_stderr, _check)

    def test_failure4(self):

        config = {
            "workers": [{
                "type":
                "router",
                "realms": [{
                    "name":
                    "realm1",
                    "roles": [{
                        "name":
                        "anonymous",
                        "permissions": [{
                            "uri": "*",
                            "publish": True,
                            "subscribe": True,
                            "call": True,
                            "register": True
                        }]
                    }]
                }],
                "transports": [{
                    "type": "web",
                    "endpoint": {
                        "type": "tcp",
                        "port": 8080
                    },
                    "paths": {
                        "/": {
                            "type": "static",
                            "directory": ".."
                        },
                        "ws": {
                            "type": "websocket"
                        }
                    }
                }]
            }, {
                "type":
                "container",
                "options": {
                    "pythonpath": [self.code_location]
                },
                "components": [{
                    "type": "class",
                    "classname": "myapp.MySession",
                    "realm": "realm1",
                    "transport": {
                        "type": "websocket",
                        "endpoint": {
                            "type": "tcp",
                            "host": "127.0.0.1",
                            "port": 8080
                        },
                        "url": "ws://127.0.0.1:8080/ws"
                    }
                }]
            }]
        }

        myapp = """
from twisted.logger import Logger
from autobahn.twisted.wamp import ApplicationSession

class MySession(ApplicationSession):

    log = Logger()

    def __init__(self, config):
        self.log.info("MySession.__init__()")
        ApplicationSession.__init__(self, config)

    def onJoin(self, details):
        self.log.info("MySession.onJoin()")
        a = 1 / 0 # trigger exception
"""

        def _check(_1, _2):
            pass

        expected_stdout = []
        expected_stderr = ["Fatal error in component", "While firing onJoin", "division by zero"]

        self._start_run(config, myapp, expected_stdout, expected_stderr, _check)

    def test_failure5(self):

        config = {
            "controller": {},
            "workers": [{
                "type":
                "router",
                "realms": [{
                    "name":
                    "realm1",
                    "roles": [{
                        "name":
                        "anonymous",
                        "permissions": [{
                            "uri": "*",
                            "publish": True,
                            "subscribe": True,
                            "call": True,
                            "register": True
                        }]
                    }]
                }],
                "transports": [{
                    "type": "web",
                    "endpoint": {
                        "type": "tcp",
                        "port": 8080
                    },
                    "paths": {
                        "/": {
                            "type": "static",
                            "directory": ".."
                        },
                        "ws": {
                            "type": "websocket"
                        }
                    }
                }]
            }, {
                "type":
                "container",
                "options": {
                    "pythonpath": [self.code_location]
                },
                "components": [{
                    "type": "class",
                    "classname": "myapp.MySession",
                    "realm": "realm1",
                    "transport": {
                        "type": "websocket",
                        "endpoint": {
                            "type": "tcp",
                            "host": "127.0.0.1",
                            "port": 8080
                        },
                        "url": "ws://127.0.0.1:8080/ws"
                    }
                }]
            }]
        }

        myapp = """
from twisted.logger import Logger
from autobahn.twisted.wamp import ApplicationSession

class MySession(ApplicationSession):

    log = Logger()

    def __init__(self, config):
        self.log.info("MySession.__init__()")
        ApplicationSession.__init__(self, config)

    def onJoin(self, details):
        self.log.info("MySession.onJoin()")
        self.leave()

    def onLeave(self, details):
        self.log.info("Session ended: {details}", details=details)
        self.disconnect()
"""

        def _check(_1, _2):
            pass

        expected_stdout = []
        expected_stderr = ["Component 'component1' failed to start; shutting down node."]

        self._start_run(config, myapp, expected_stdout, expected_stderr, _check)

    def test_failure6(self):

        config = {
            "controller": {},
            "workers": [{
                "type":
                "router",
                "realms": [{
                    "name":
                    "realm1",
                    "roles": [{
                        "name":
                        "anonymous",
                        "permissions": [{
                            "uri": "*",
                            "publish": True,
                            "subscribe": True,
                            "call": True,
                            "register": True
                        }]
                    }]
                }],
                "transports": [{
                    "type": "web",
                    "endpoint": {
                        "type": "tcp",
                        "port": 8080
                    },
                    "paths": {
                        "/": {
                            "type": "static",
                            "directory": ".."
                        },
                        "ws": {
                            "type": "websocket"
                        }
                    }
                }]
            }, {
                "type":
                "container",
                "options": {
                    "pythonpath": [self.code_location]
                },
                "components": [{
                    "type": "class",
                    "classname": "myapp.MySession",
                    "realm": "realm1",
                    "transport": {
                        "type": "websocket",
                        "endpoint": {
                            "type": "tcp",
                            "host": "127.0.0.1",
                            "port": 8080
                        },
                        "url": "ws://127.0.0.1:8080/ws"
                    }
                }]
            }]
        }

        myapp = """
from twisted.logger import Logger
from twisted.internet.defer import inlineCallbacks
from autobahn.twisted.wamp import ApplicationSession
from autobahn.twisted.util import sleep

class MySession(ApplicationSession):

    log = Logger()

    def __init__(self, config):
        self.log.info("MySession.__init__()")
        ApplicationSession.__init__(self, config)

    @inlineCallbacks
    def onJoin(self, details):
        self.log.info("MySession.onJoin()")
        self.log.info("Sleeping a couple of secs and then shutting down ..")
        yield sleep(2)
        self.leave()

    def onLeave(self, details):
        self.log.info("Session ended: {details}", details=details)
        self.disconnect()

"""

        def _check(_1, _2):
            pass

        expected_stdout = [
            "Session ended: CloseDetails", "Sleeping a couple of secs and then shutting down",
            "Container is hosting no more components: shutting down"
        ]
        expected_stderr = []

        self._start_run(config, myapp, expected_stdout, expected_stderr, _check)

    def test_failure7(self):

        config = {
            "workers": [{
                "type":
                "router",
                "realms": [{
                    "name":
                    "realm1",
                    "roles": [{
                        "name":
                        "anonymous",
                        "permissions": [{
                            "uri": "*",
                            "publish": True,
                            "subscribe": True,
                            "call": True,
                            "register": True
                        }]
                    }]
                }],
                "transports": [{
                    "type": "web",
                    "endpoint": {
                        "type": "tcp",
                        "port": 8080
                    },
                    "paths": {
                        "/": {
                            "type": "static",
                            "directory": ".."
                        },
                        "ws": {
                            "type": "websocket"
                        }
                    }
                }]
            }, {
                "type":
                "container",
                "options": {
                    "pythonpath": [self.code_location]
                },
                "components": [{
                    "type": "class",
                    "classname": "myapp.MySession",
                    "realm": "realm1",
                    "transport": {
                        "type": "websocket",
                        "endpoint": {
                            "type": "tcp",
                            "host": "127.0.0.1",
                            "port": 8090
                        },
                        "url": "ws://127.0.0.1:8090/ws"
                    }
                }]
            }]
        }

        myapp = """
from twisted.logger import Logger
from autobahn.twisted.wamp import ApplicationSession

class MySession(ApplicationSession):

    log = Logger()

    def __init__(self, config):
        self.log.info("MySession.__init__()")
        ApplicationSession.__init__(self, config)

    def onJoin(self, details):
        self.log.info("MySession.onJoin()")
        self.leave()
"""

        def _check(_1, _2):
            pass

        expected_stdout = []
        expected_stderr = [("Could not connect container component to router - transport " "establishment failed")]

        self._start_run(config, myapp, expected_stdout, expected_stderr, _check)


class InitTests(CLITestBase):
    def test_hello(self):
        def _check(lc, reactor):
            if "published to 'oncounter'" in self.stdout.getvalue():
                lc.stop()
                try:
                    reactor.stop()
                except:
                    pass

        appdir = self.mktemp()
        cbdir = os.path.join(appdir, ".crossbar")

        reactor = SelectReactor()
        main.main("crossbar", ["init", "--appdir={}".format(appdir), "--template=hello:python"], reactor=reactor)

        self.assertIn("Application template initialized", self.stdout.getvalue())

        reactor = SelectReactor()
        make_lc(self, reactor, _check)

        # In case it hard-locks
        reactor.callLater(self._subprocess_timeout, reactor.stop)

        main.main("crossbar", ["start", "--cbdir={}".format(cbdir.path), "--logformat=syslogd"], reactor=reactor)

        stdout_expected = ["published to 'oncounter'"]

        for i in stdout_expected:
            self.assertIn(i, self.stdout.getvalue())


if not os.environ.get("CB_FULLTESTS"):
    del ContainerRunningTests
    del InitTests



###############################################################################
#
# Crossbar.io Shell
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import os
import sys
import argparse
import binascii

import txaio
from autobahn.wamp import cryptosign
from autobahn.wamp.exception import ApplicationError
from autobahn.websocket.util import parse_url
from autobahn.twisted.wamp import ApplicationSession, ApplicationRunner

__all__ = ('ShellClient', 'ManagementClientSession', 'run', 'create_management_session')

from crossbar.shell.command import CmdListManagementRealms
from twisted.internet.defer import Deferred


class ShellClient(ApplicationSession):
    """
    Management CLI application session.
    """
    def onConnect(self):  # noqa: N802
        self.log.info('{klass}.onConnect()', klass=self.__class__.__name__)

        self._key = self.config.extra['key']
        self._command = self.config.extra.get('command', None)
        self._main = self.config.extra.get('main', None)

        # authentication extra information for wamp-cryptosign
        #
        extra = {
            # forward the client pubkey: this allows us to omit authid as
            # the router can identify us with the pubkey already
            'pubkey': self._key.public_key(),

            # not yet implemented. a public key the router should provide
            # a trustchain for it's public key. the trustroot can eg be
            # hard-coded in the client, or come from a command line option.
            'trustroot': self.config.extra.get('trustroot', None),

            # not yet implemented. for authenticating the router, this
            # challenge will need to be signed by the router and send back
            # in AUTHENTICATE for client to verify. A string with a hex
            # encoded 32 bytes random value.
            'challenge': self.config.extra.get('challenge', None),

            # use TLS channel binding
            'channel_binding': self.config.extra.get('channel_binding', None),
        }

        # used for user login/registration activation code
        for k in ['activation_code', 'request_new_activation_code']:
            if k in self.config.extra and self.config.extra[k]:
                extra[k] = self.config.extra[k]

        # now request to join ..
        self.join(self.config.realm,
                  authmethods=['cryptosign'],
                  authid=self.config.extra.get('authid', None),
                  authrole=self.config.extra.get('authrole', None),
                  authextra=extra)

    def onChallenge(self, challenge):  # noqa: N802
        self.log.info('{klass}.onChallenge(challenge={challenge})', klass=self.__class__.__name__, challenge=challenge)

        # sign and send back the challenge with our private key.
        try:
            sig = self._key.sign_challenge(self, challenge)
        except Exception as e:
            self.log.failure()
            self.leave(ApplicationError.AUTHENTICATION_FAILED, str(e))
        else:
            return sig

    async def onJoin(self, details):  # noqa: N802
        self.log.info('{klass}.onJoin(details={details})', klass=self.__class__.__name__, details=details)

        done = self.config.extra.get('done', None)

        result = None
        error = None
        if self._command:
            self.log.info('{klass}: running command {command}', klass=self.__class__.__name__, command=self._command)
            try:
                result = await self._command.run(self)
                self.log.info('command run with result {result}', result=result)
            except Exception as e:
                self.log.warn('command failed: {error}', error=e)
                error = e
        elif self._main:
            self.log.info('{klass}: running main function {main}', klass=self.__class__.__name__, main=self._main)
            try:
                result = await self._main(self)
                self.log.info('main run with result {result}', result=result)
            except Exception as e:
                self.log.warn('main failed: {error}', error=e)
                error = e
        else:
            self.log.info('{klass}: no command or main function to run!', klass=self.__class__.__name__)

        if done and not txaio.is_called(done):
            if error:
                self.log.warn('{klass}: command returned with error ({error})',
                              klass=self.__class__.__name__,
                              error=error)
                txaio.reject(done, error)
            else:
                self.log.info('{klass}: command returned with success ({result})',
                              klass=self.__class__.__name__,
                              result=result)
                txaio.resolve(done, (details, result))

        self.log.info('{klass}.onJoin(): finished!', klass=self.__class__.__name__)

        if self._main:
            self.leave()

    def onLeave(self, details):  # noqa: N802
        self.log.info('{klass}.onLeave(details={details})', klass=self.__class__.__name__, details=details)

        # reason=<wamp.error.authentication_failed>
        if details.reason != 'wamp.close.normal':
            done = self.config.extra.get('done', None)
            error = ApplicationError(details.reason, details.message)
            if done and not txaio.is_called(done):
                self.log.warn('{klass}: command returned with error ({error})',
                              klass=self.__class__.__name__,
                              error=error)
                txaio.reject(done, error)

        self.disconnect()

    def onDisconnect(self):  # noqa: N802
        self.log.info('{klass}.onDisconnect()', klass=self.__class__.__name__)
        if self._main:
            try:
                self.config.runner.stop()
                self.disconnect()
            except:
                self.log.failure()
            from twisted.internet import reactor
            if reactor.running:
                reactor.stop()


class ManagementClientSession(ApplicationSession):
    def onConnect(self):
        self._key = self.config.extra['key']
        extra = {
            'pubkey': self._key.public_key(),
            'trustroot': self.config.extra.get('trustroot', None),
            'challenge': self.config.extra.get('challenge', None),
            'channel_binding': self.config.extra.get('channel_binding', None),
        }
        for k in ['activation_code', 'request_new_activation_code']:
            if k in self.config.extra and self.config.extra[k]:
                extra[k] = self.config.extra[k]

        self.join(self.config.realm,
                  authmethods=['cryptosign'],
                  authid=self.config.extra.get('authid', None),
                  authrole=self.config.extra.get('authrole', None),
                  authextra=extra)

    def onChallenge(self, challenge):
        return self._key.sign_challenge(self, challenge)

    def onJoin(self, details):
        if 'ready' in self.config.extra:
            self.config.extra['ready'].callback((self, details))

    def onLeave(self, reason):
        self.disconnect()


def create_management_session(url='wss://master.xbr.network/ws',
                              realm='com.crossbario.fabric',
                              privkey_file='default.priv'):

    txaio.start_logging(level='info')

    privkey_file = os.path.join(os.path.expanduser('~/.crossbar'), privkey_file)

    # for authenticating the management client, we need a Ed25519 public/private key pair
    # here, we are reusing the user key - so this needs to exist before
    privkey_hex = None
    user_id = None

    if not os.path.exists(privkey_file):
        raise Exception('private key file {} does not exist'.format(privkey_file))
    else:
        with open(privkey_file, 'r') as f:
            data = f.read()
            for line in data.splitlines():
                if line.startswith('private-key-ed25519'):
                    privkey_hex = line.split(':')[1].strip()
                if line.startswith('user-id'):
                    user_id = line.split(':')[1].strip()

    if privkey_hex is None:
        raise Exception('no private key found in keyfile!')

    if user_id is None:
        raise Exception('no user ID found in keyfile!')

    url_is_secure, _, _, _, _, _ = parse_url(url)

    key = cryptosign.SigningKey.from_key_bytes(binascii.a2b_hex(privkey_hex))
    extra = {
        'key': key,
        'authid': user_id,
        'ready': Deferred(),
        'return_code': None,
        'command': CmdListManagementRealms(),

        # WAMP-cryptosign TLS channel binding
        'channel_binding': 'tls-unique' if url_is_secure else None,
    }

    runner = ApplicationRunner(url=url, realm=realm, extra=extra)
    runner.run(ManagementClientSession, start_reactor=False)

    return extra['ready']


def run(main=None, parser=None):

    # parse command line arguments
    parser = parser or argparse.ArgumentParser()

    parser.add_argument('--debug',
                        dest='debug',
                        action='store_true',
                        default=False,
                        help='Enable logging at level "debug".')
    parser.add_argument('--url',
                        dest='url',
                        type=str,
                        default='wss://master.xbr.network/ws',
                        help='Management service of the XBR Network'
                        '(default: wss://master.xbr.network/ws')
    parser.add_argument('--realm',
                        dest='realm',
                        type=str,
                        default=None,
                        help='The (management) realm to join on the management server')
    parser.add_argument('--keyfile',
                        dest='keyfile',
                        type=str,
                        default=None,
                        help='The private client key file to use for authentication.')
    parser.add_argument('--authmethod',
                        dest='authmethod',
                        type=str,
                        default='cryptosign',
                        help='Authentication method: cryptosign or anonymous')

    args = parser.parse_args()

    if args.debug:
        txaio.start_logging(level='debug')
    else:
        txaio.start_logging(level='info')

    args.keyfile = os.path.abspath(os.path.expanduser(args.keyfile))
    print('usering keyfile from', args.keyfile)

    extra = None
    if args.authmethod == 'cryptosign':

        # for authenticating the management client, we need a Ed25519 public/private key pair
        # here, we are reusing the user key - so this needs to exist before
        privkey_file = os.path.expanduser(args.keyfile)
        privkey_hex = None
        user_id = None

        if not os.path.exists(privkey_file):
            raise Exception('private key file {} does not exist'.format(privkey_file))
        else:
            with open(privkey_file, 'r') as f:
                data = f.read()
                for line in data.splitlines():
                    if line.startswith('private-key-ed25519'):
                        privkey_hex = line.split(':')[1].strip()
                    if line.startswith('user-id'):
                        user_id = line.split(':')[1].strip()

        if privkey_hex is None:
            raise Exception('no private key found in keyfile!')

        if user_id is None:
            raise Exception('no user ID found in keyfile!')

        key = cryptosign.SigningKey.from_key_bytes(binascii.a2b_hex(privkey_hex))

        extra = {'args': args, 'key': key, 'authid': user_id, 'main': main, 'return_code': None}

    elif args.authmethod == 'anonymous':

        extra = {'args': args, 'main': main, 'return_code': None}

    else:
        raise Exception('logic error')

    runner = ApplicationRunner(url=args.url, realm=args.realm, extra=extra)

    runner.run(ShellClient)

    return_code = extra['return_code']
    if isinstance(return_code, int) and return_code != 0:
        sys.exit(return_code)


if __name__ == '__main__':
    run()



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import os
import binascii
from pprint import pformat
from typing import Optional, Union, Dict, Any

import nacl
from nacl.signing import VerifyKey
from nacl.exceptions import BadSignatureError

from twisted.internet.defer import Deferred

import txaio

from autobahn import util
from autobahn.util import hltype, hlid, hlval
from autobahn.wamp.types import Accept, Deny, HelloDetails, Challenge, TransportDetails
from autobahn.wamp.exception import ApplicationError

from crossbar.router.auth.pending import PendingAuth
from crossbar.interfaces import IRealmContainer, IPendingAuth

__all__ = (
    'PendingAuthCryptosign',
    'PendingAuthCryptosignProxy',
)


class PendingAuthCryptosign(PendingAuth):
    """
    Pending Cryptosign authentication.
    """

    log = txaio.make_logger()

    AUTHMETHOD = 'cryptosign'

    def __init__(self, pending_session_id: int, transport_details: TransportDetails, realm_container: IRealmContainer,
                 config: Dict[str, Any]):
        super(PendingAuthCryptosign, self).__init__(
            pending_session_id,
            transport_details,
            realm_container,
            config,
        )
        # https://tools.ietf.org/html/rfc5056
        # https://tools.ietf.org/html/rfc5929
        # https://www.ietf.org/proceedings/90/slides/slides-90-uta-0.pdf
        self._channel_id = transport_details.channel_id.get('tls-unique',
                                                            None) if transport_details.channel_id else None

        self._verify_key: Optional[VerifyKey] = None
        self._challenge: Optional[bytes] = None
        self._expected_signed_message: Optional[bytes] = None

        # create a map: pubkey -> authid
        # this is to allow clients to authenticate without specifying an authid
        if config['type'] == 'static':
            self._pubkey_to_authid = {}
            for authid, principal in self._config.get('principals', {}).items():
                for pubkey in principal['authorized_keys']:
                    self._pubkey_to_authid[pubkey] = authid

    def _compute_challenge(self, requested_channel_binding: Optional[str]) -> Dict[str, Any]:
        self._challenge = os.urandom(32)
        if self._channel_id and requested_channel_binding == 'tls-unique':
            self._expected_signed_message = util.xor(self._challenge, self._channel_id)
        else:
            self._expected_signed_message = self._challenge

        extra = {
            'challenge': binascii.b2a_hex(self._challenge).decode(),
            'channel_binding': requested_channel_binding,
        }
        self.log.info(
            '{func}::_compute_challenge(channel_binding={channel_binding})[channel_id={channel_id}] -> extra=\n{extra}',
            func=hltype(self.hello),
            channel_id=hlid('0x' + binascii.b2a_hex(self._channel_id).decode()) if self._channel_id else None,
            channel_binding=hlval('"' + requested_channel_binding +
                                  '"') if requested_channel_binding is not None else None,
            extra=pformat(extra))
        return extra

    def hello(self, realm: str, details: HelloDetails) -> Union[Accept, Deny, Challenge]:
        self.log.info('{func}::hello(realm="{realm}", details.authid="{authid}", details.authrole="{authrole}")',
                      func=hltype(self.hello),
                      realm=hlid(realm),
                      authid=hlid(details.authid),
                      authrole=hlid(details.authrole))

        # the channel binding requested by the client authenticating
        requested_channel_binding = details.authextra.get('channel_binding', None) if details.authextra else None
        if requested_channel_binding is not None and requested_channel_binding not in ['tls-unique']:
            return Deny(message='invalid channel binding type "{}" requested'.format(requested_channel_binding))
        else:
            self.log.info(
                "WAMP-cryptosign CHANNEL BINDING requested: channel_binding={channel_binding}, channel_id={channel_id}",
                channel_binding=requested_channel_binding,
                channel_id=self._channel_id)

        # remember the realm the client requested to join (if any)
        self._realm = realm

        # remember the authid the client wants to identify as (if any)
        self._authid = details.authid

        # use static principal database from configuration
        if self._config['type'] == 'static':

            self._authprovider = 'static'

            # get client's pubkey, if it was provided in authextra
            pubkey = None
            if details.authextra and 'pubkey' in details.authextra:
                pubkey = details.authextra['pubkey']

            # if the client provides it's public key, that's enough to identify,
            # and we can infer the authid from that. BUT: that requires that
            # there is a 1:1 relation between authid's and pubkey's !! see below (*)
            if self._authid is None:
                if pubkey:
                    # we do a naive search, but that is ok, since "static mode" is from
                    # node configuration, and won't contain a lot principals anyway
                    for _authid, _principal in self._config.get('principals', {}).items():
                        if pubkey in _principal['authorized_keys']:
                            # (*): this is necessary to detect multiple authid's having the same pubkey
                            # in which case we couldn't reliably map the authid from the pubkey
                            if self._authid is None:
                                self._authid = _authid
                            else:
                                return Deny(message='cannot infer client identity from pubkey: multiple authids '
                                            'in principal database have this pubkey')
                    if self._authid is None:
                        return Deny(message='cannot identify client: no authid requested and no principal found '
                                    'for provided extra.pubkey')
                else:
                    return Deny(message='cannot identify client: no authid requested and no extra.pubkey provided')

            principals = self._config.get('principals', {})
            if self._authid in principals:

                principal = principals[self._authid]

                if pubkey and (pubkey not in principal['authorized_keys']):
                    self.log.warn(
                        'extra.pubkey {pubkey} provided does not match any one of authorized_keys for the principal [func="{func}"]:\n{principals}',
                        func=hltype(self.hello),
                        realm=hlid(realm),
                        authid=hlid(details.authid),
                        pubkey=hlval(pubkey),
                        principals=pformat(principals))
                    return Deny(
                        message='extra.pubkey provided does not match any one of authorized_keys for the principal')

                error = self._assign_principal(principal)
                if error:
                    return error

                self._verify_key = VerifyKey(pubkey, encoder=nacl.encoding.HexEncoder)

                extra = self._compute_challenge(requested_channel_binding)
                return Challenge(self._authmethod, extra)

            else:
                self.log.warn(
                    'no principal with authid "{authid}" exists in principals for realm "{realm}" [func="{func}"]:\n{principals}',
                    func=hltype(self.hello),
                    realm=hlid(realm),
                    authid=hlid(self._authid),
                    principals=pformat(principals))
                return Deny(message='no principal with authid "{}" exists'.format(self._authid))

        elif self._config['type'] == 'dynamic':

            self._authprovider = 'dynamic'

            d = Deferred()

            d1 = txaio.as_future(self._init_dynamic_authenticator)

            def initialized(error=None):
                if error:
                    d.errback(error)
                    return

                self._session_details['authmethod'] = self._authmethod  # from AUTHMETHOD, via base
                self._session_details['authid'] = details.authid
                self._session_details['authrole'] = details.authrole
                self._session_details['authextra'] = details.authextra

                self.log.debug(
                    'Calling dynamic authenticator [proc="{proc}", realm="{realm}", session={session}, authid="{authid}", authrole="{authrole}"]',
                    proc=self._authenticator,
                    realm=self._authenticator_session._realm,
                    session=self._authenticator_session._session_id,
                    authid=self._authenticator_session._authid,
                    authrole=self._authenticator_session._authrole)

                d2 = self._authenticator_session.call(self._authenticator, realm, details.authid,
                                                      self._session_details)

                def on_authenticate_ok(principal):
                    self.log.debug(
                        '{klass}.hello(realm="{realm}", details={details}) -> on_authenticate_ok(principal={principal})',
                        klass=self.__class__.__name__,
                        realm=realm,
                        details=details,
                        principal=principal)
                    _error = self._assign_principal(principal)
                    if _error:
                        d.callback(_error)
                        return

                    self._verify_key = VerifyKey(principal['pubkey'], encoder=nacl.encoding.HexEncoder)

                    extra = self._compute_challenge(requested_channel_binding)
                    d.callback(Challenge(self._authmethod, extra))

                def on_authenticate_error(_error):
                    self.log.debug(
                        '{klass}.hello(realm="{realm}", details={details}) -> on_authenticate_error(error={error})',
                        klass=self.__class__.__name__,
                        realm=realm,
                        details=details,
                        error=_error)
                    try:
                        d.callback(self._marshal_dynamic_authenticator_error(_error))
                    except:
                        self.log.failure()
                        d.callback(_error)

                d2.addCallbacks(on_authenticate_ok, on_authenticate_error)
                return d2

            def initialized_error(fail):
                self.log.failure('Internal error (3): {log_failure.value}', failure=fail)
                d.errback(fail)

            d1.addCallbacks(initialized, initialized_error)

            return d

        elif self._config['type'] == 'function':
            self._authprovider = 'function'

            init_d = txaio.as_future(self._init_function_authenticator)

            def init(_error):
                if _error:
                    return _error

                self._session_details['authmethod'] = self._authmethod  # from AUTHMETHOD, via base
                self._session_details['authid'] = details.authid
                self._session_details['authrole'] = details.authrole
                self._session_details['authextra'] = details.authextra

                auth_d = txaio.as_future(self._authenticator, realm, details.authid, self._session_details)

                def on_authenticate_ok(principal):
                    self.log.debug(
                        '{klass}.hello(realm="{realm}", details={details}) -> on_authenticate_ok(principal={principal})',
                        klass=self.__class__.__name__,
                        realm=realm,
                        details=details,
                        principal=principal)
                    _error = self._assign_principal(principal)
                    if _error:
                        return _error

                    self._verify_key = VerifyKey(principal['pubkey'], encoder=nacl.encoding.HexEncoder)

                    _extra = self._compute_challenge(requested_channel_binding)
                    return Challenge(self._authmethod, _extra)

                def on_authenticate_error(err):
                    self.log.warn(
                        '{klass}.hello(realm="{realm}", details={details}) -> on_authenticate_error(err={err})',
                        klass=self.__class__.__name__,
                        realm=realm,
                        details=details,
                        err=err)
                    try:
                        return self._marshal_dynamic_authenticator_error(err)
                    except Exception as e:
                        _error = ApplicationError.AUTHENTICATION_FAILED
                        message = 'marshalling of function-based authenticator error return failed: {}'.format(e)
                        self.log.warn('{klass}.hello.on_authenticate_error() - {msg}', msg=message)
                        return Deny(_error, message)

                auth_d.addCallbacks(on_authenticate_ok, on_authenticate_error)
                return auth_d

            init_d.addBoth(init)
            return init_d

        else:
            # should not arrive here, as config errors should be caught earlier
            return Deny(message='invalid authentication configuration (authentication type "{}" is unknown)'.format(
                self._config['type']))

    def authenticate(self, signature: str) -> Union[Accept, Deny]:
        """
        Verify the signed message sent by the client. With WAMP-cryptosign, this must be 96 bytes (as a string
        in HEX encoding): the concatenation of the Ed25519 signature (64 bytes) and the 32 bytes we sent
        as a challenge previously, XORed with the 32 bytes transport channel ID (if available).
        """
        try:
            if not isinstance(signature, str):
                return Deny(message='invalid type {} for signed message'.format(type(signature)))

            try:
                signed_message = binascii.a2b_hex(signature)
            except TypeError:
                return Deny(message='signed message is invalid (not a HEX encoded string)')

            if len(signed_message) != 96:
                return Deny(message='signed message has invalid length (was {}, but should have been 96)'.format(
                    len(signed_message)))

            # now verify the signed message versus the client public key
            assert self._verify_key
            try:
                message = self._verify_key.verify(signed_message)
            except BadSignatureError:
                return Deny(message='signed message has invalid signature')

            # and check that the message signed by the client is really what we expect
            assert self._expected_signed_message
            if message != self._expected_signed_message:
                return Deny(message='message signed is bogus [got 0x{}, expected 0x{}]'.format(
                    binascii.b2a_hex(message).decode(),
                    binascii.b2a_hex(self._expected_signed_message).decode()))

            # signature was valid _and_ the message that was signed is equal to
            # what we expected => accept the client
            return self._accept()

        # should not arrive here, but who knows
        except Exception as e:
            self.log.failure()
            return Deny(message='INTERNAL ERROR ({})'.format(e))


class PendingAuthCryptosignProxy(PendingAuthCryptosign):
    """
    Pending Cryptosign authentication with additions for proxy
    """

    log = txaio.make_logger()
    AUTHMETHOD = 'cryptosign-proxy'

    def hello(self, realm, details):
        self.log.debug('{klass}.hello(realm={realm}, details={details}) ...',
                       klass=self.__class__.__name__,
                       realm=realm,
                       details=details)
        if not details.authextra:
            return Deny(message='missing required details.authextra')
        for attr in ['proxy_authid', 'proxy_authrole', 'proxy_realm']:
            if attr not in details.authextra:
                return Deny(message='missing required attribute {} in details.authextra'.format(attr))

        if details.authrole is None:
            details.authrole = details.authextra.get('proxy_authrole', None)
        if details.authid is None:
            details.authid = details.authextra.get('proxy_authid', None)

        # with authenticators of type "*-proxy", the principal returned in authenticating the
        # incoming backend connection is ignored ..
        f = txaio.as_future(super(PendingAuthCryptosignProxy, self).hello, realm, details)

        def assign(res):
            if isinstance(res, Deny):
                return res

            # the incoming backend connection from the proxy frontend is authenticated as the principal
            # the frontend proxy has _already_ authenticated the actual client (before even connecting and
            # authenticating to the backend here)
            principal = {
                'realm': details.authextra['proxy_realm'],
                'authid': details.authextra['proxy_authid'],
                'role': details.authextra['proxy_authrole'],

                # the authextra intended for the principal is forwarded from the proxy
                'extra': details.authextra.get('proxy_authextra', None)
            }
            self._assign_principal(principal)

            self.log.debug(
                '{klass}.hello(realm={realm}, details={details}) -> principal={principal}',
                klass=self.__class__.__name__,
                realm=realm,
                details=details,
                principal=principal,
            )
            return self._accept()

        def error(_err):
            return Deny("Internal error: {}".format(_err))

        txaio.add_callbacks(f, assign, error)
        return f


IPendingAuth.register(PendingAuthCryptosign)



###############################################################################
#
# Crossbar.io Quickstart
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import os
import uuid
import six

import click
import yaml
from cookiecutter.main import cookiecutter

from autobahn import util

README = """# Crossbar.io Quickstart Project

This project was generated using quickstart templates for Crossbar.io based services.

The services are [Docker](https://www.docker.com/) based, and the service bundling is
using [Docker Compose](https://docs.docker.com/compose/).

To build your services:

```console
make build
```

To check the service versions:

```console
make version
```

To start your services:

```console
make start
```

> Note: the latter is simply starting all services with `docker-compose up`. To start
the services in the background, type `docker-compose up --detach`.
"""

MAKEFILE = """
SUBDIRS = $(shell ls -d */)

default: start

version:
\tfor dir in $(SUBDIRS) ; do \
\t\techo "$$dir" ; \
\t\tcat $$dir/service.json ; \
\t\tmake -C  $$dir version ; \
\tdone

build:
\tdocker-compose build

# the following will automatically build the images from the Dockerfiles
# references in docker-compose.yml, and it will also rebuild those images
# and recreate containers as needed (eg when the Dockerfile or again a file
# referenced therein, such as your actual application source code files, changes)
#
start:
\tdocker-compose up --build
"""

readme_filename = 'README.md'
makefile_filename = 'Makefile'
docker_compose_filename = 'docker-compose.yml'

DEVMODE = False

if DEVMODE:
    CC1 = '/home/oberstet/scm/crossbario'
    CC2 = '/home/oberstet/scm/xbr'
else:
    CC1 = 'gh:crossbario'
    CC2 = 'gh:xbr'

# built-in cookiecutters.
#
_cookiecutters = [
    # Crossbar.io
    (None, 'Crossbar.io'),
    ('{}/cookiecutter-crossbar'.format(CC1), 'Create a Crossbar.io OSS app router'),
    ('{}/cookiecutter-crossbar-fabric'.format(CC1), 'Create a Crossbar.io app router'),
    ('{}/cookiecutter-crossbar-fabric-center'.format(CC1), 'Create a Crossbar.io cluster controller'),

    # XBR
    (None, 'XBR'),
    ('{}/cookiecutter-xbr-api'.format(CC2), 'Create a XBR API bundle'),
    ('{}/cookiecutter-crossbar-fabric-xbr'.format(CC1), 'Create a Crossbar.io XBR data market'),

    # Autobahn
    (None, 'Autobahn'),
    ('{}/cookiecutter-autobahn-python'.format(CC1), 'Create a Python based app or XBR service'),
    ('{}/cookiecutter-autobahn-js'.format(CC1), 'Create a JavaScript based app or XBR service'),
    ('{}/cookiecutter-autobahn-java'.format(CC1), 'Create a Java based app or XBR service'),
    ('{}/cookiecutter-autobahn-cpp'.format(CC1), 'Create a C++ based app or XBR service'),

    # Community project
    (None, 'Community'),
    ('{}/cookiecutter-wampsharp'.format(CC1), 'Create a WampSharp/C# based app service'),
    ('{}/cookiecutter-nexus-go'.format(CC1), 'Create a Nexus/Go based app service'),
]


def hl(text, color='yellow', bold=True):
    if not isinstance(text, six.text_type):
        text = '{}'.format(text)
    return click.style(text, fg=color, bold=bold)


def _initialize():
    if not os.path.isfile(readme_filename):
        with open(readme_filename, 'w') as fd:
            fd.write(README)
        click.echo('{} created'.format(readme_filename))

    if not os.path.isfile(makefile_filename):
        with open(makefile_filename, 'w') as fd:
            fd.write(MAKEFILE)
        click.echo('{} created'.format(makefile_filename))

    docker_compose_filename = 'docker-compose.yml'
    if not os.path.isfile(docker_compose_filename):
        docker_compose = {'version': '3.0', 'services': {}}
        with open(docker_compose_filename, 'w') as fd:
            fd.write(yaml.safe_dump(docker_compose))
        click.echo('{} created'.format(docker_compose_filename))

    docker_compose = None
    with open(docker_compose_filename) as fd:
        data = fd.read()
        docker_compose = yaml.safe_load(data)

    if not isinstance(docker_compose, dict):
        raise click.ClickException('invalid type {} found in {} for top level object'.format(
            type(docker_compose), docker_compose_filename))

    if 'services' not in docker_compose:
        raise click.ClickException(
            'no services attribute found in top level object for {}'.format(docker_compose_filename))

    return docker_compose


def run():
    click.echo('\n{cb} Project Quickstart\n'.format(cb=hl('Crossbar.io / XBR')))

    _templates = {}

    num = 0
    click.echo('  {}: Exit'.format(hl(num)))

    for template, desc in _cookiecutters:
        if template:
            num += 1
            _templates[num] = template
            template_disp = hl(template, bold=False)
            click.echo('  {num}: {desc:46s} [{template}]'.format(num=hl(str(num).ljust(3)),
                                                                 desc=desc,
                                                                 template=template_disp))
        else:
            click.echo('\n {}:\n'.format(hl(desc)))
    click.echo('')

    select = None
    while select not in range(0, num + 1):
        select = click.prompt('Please select', type=int, default='0')

    if select > 0:

        template = _templates[select]
        click.echo('Initializing cookiecutter {} ...'.format(template))

        # initialize services root directory and return contents of
        # docker-compose.py as object
        docker_compose = _initialize()

        extra_context = {
            'uid': os.getuid(),
            'service_uuid': str(uuid.uuid4()),
            'generated': util.utcnow(),
        }
        output_dir = '.'

        # cookiecutter returns the fully qualified path within which the template
        # was initialized.
        output_dir = cookiecutter(template, output_dir=output_dir, extra_context=extra_context)

        # the last part of the fully qualified output directory is the service name
        # that comes from "cookiecutter.json"
        service_name = os.path.basename(output_dir)

        # we expect the cookiecutter to produce a docker-compose-<service_name>.yml file
        service_docker_compose_filename = os.path.join(output_dir, 'docker-compose-{}.yml'.format(service_name))

        if not os.path.isfile(service_docker_compose_filename):
            raise click.ClickException(
                'docker-compose fragment for service was not generated by cookiecutter. missing file:\n{}'.format(
                    service_docker_compose_filename))

        # now load the docker-compose service fragment from the file generated by cookiecutter
        service_docker_compose = None
        with open(service_docker_compose_filename) as fd:
            data = fd.read()
            service_docker_compose = yaml.safe_load(data)

        # update the docker-compose object
        if service_name in docker_compose['services']:
            click.echo('updating service "{}" existing in docker-compose.yml'.format(service_name))
        else:
            click.echo('adding service "{}" to docker-compose.yml'.format(service_name))

        docker_compose['services'][service_name] = service_docker_compose

        # write out the docker-compose object to the file
        with open(docker_compose_filename, 'w') as fd:
            fd.write(yaml.safe_dump(docker_compose))
        click.echo('{} written'.format(docker_compose_filename))



# coding=utf8

##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import os
import sys
import uuid
import mimetypes
import argparse
import binascii

from twisted.python import log
from twisted.internet import reactor
from twisted.web.server import Site
from twisted.web.wsgi import WSGIResource

from flask import Flask, request, session, render_template, redirect

import numpy as np
from txaio import time_ns
from autobahn.util import generate_activation_code

from cfxdb.xbrnetwork import Account, VerifiedAction

from validate_email import validate_email

from ._backend import is_valid_username
from ._mailgw import MailGateway

# make sure we have MIME types defined for all the
# file types we use
mimetypes.add_type('image/svg+xml', '.svg')
mimetypes.add_type('text/javascript', '.jgz')

# Jinja2 extension for Pygments
#
# Note: To generate Pygments CSS file for style: pygmentize -S default -f html > pygments.css
import jinja2_highlight  # noqa


class SiteFlask(Flask):
    jinja_options = dict(Flask.jinja_options)
    jinja_options.setdefault('extensions', []).append('jinja2_highlight.HighlightExtension')


HAS_HIGHLIGHT = True

# FIXME
_USERNAME_PAT_STR = None

# Main app object
#
app = SiteFlask(__name__)
app.secret_key = str(uuid.uuid4())

app.config['DEBUG'] = True
app.config['WEBSITE_URL'] = 'http://localhost:8090'
# app.config['WEBSITE_URL'] = 'https://xbr.network'
app.config['MAILGUN'] = MailGateway(os.environ.get('MAILGUN_URL', None), os.environ.get('MAILGUN_KEY', None),
                                    os.environ.get('MAILGUN_FROM', 'XBR Network <no-reply@mailing.xbr.network>'),
                                    app.config['WEBSITE_URL'])


@app.route('/')
def page_home():
    session['site_area'] = 'landing'
    session['site-page'] = None
    return render_template('index.html')


@app.route('/privacy/')
def page_privacy():
    session['site_area'] = 'legal'
    session['site-page'] = None
    return render_template('privacy.html')


@app.route('/impressum/')
def page_impressum():
    session['site_area'] = 'legal'
    session['site-page'] = None
    return render_template('anbieterkennung.html')


@app.route('/industrial/')
def page_industrial():
    session['site_area'] = 'sectors'
    session['site-page'] = None
    return render_template('sector_industrial.html')


@app.route('/automotive/')
def page_automotive():
    session['site_area'] = 'sectors'
    session['site-page'] = None
    return render_template('sector_automotive.html')


@app.route('/onboard/')
def page_xbr_onboard():
    session['site_area'] = 'landing'
    session['site-page'] = None
    return render_template('xbr_onboard.html')


@app.route('/submit-onboard', methods=['POST'])
def page_xbr_submit_onboard():
    session['site_area'] = 'landing'
    session['site-page'] = None
    # ImmutableMultiDict([
    #   ('onboard_member_name', 'oberstet'),
    #   ('onboard_member_email', 'tobias.oberstein@crossbario.com'),
    #   ('onboard_wallet_type', 'imported'),
    #   ('onboard_wallet_address', '0x6231eECbA6e7983efe5ce6d16972E16cCcD97CE7'),
    #   ('onboard_accept_eula', 'on')
    # ])
    onboard_member_name = request.form.get('onboard_member_name', None)
    onboard_member_email = request.form.get('onboard_member_email', None)
    onboard_wallet_type = request.form.get('onboard_wallet_type', None)
    onboard_wallet_address = request.form.get('onboard_wallet_address', None)
    onboard_accept_eula = request.form.get('onboard_accept_eula', None)

    print('page_xbr_submit_onboard:')
    print('  onboard_member_name', onboard_member_name)
    print('  onboard_member_email', onboard_member_email)
    print('  onboard_wallet_type', onboard_wallet_type)
    print('  onboard_wallet_address', onboard_wallet_address)
    print('  onboard_accept_eula', onboard_accept_eula)

    if onboard_wallet_type not in Account.WALLET_TYPE_FROM_STRING:
        return render_template('xbr_onboard_submit_error.html',
                               onboard_member_error='Invalid wallet type "{}"'.format(onboard_wallet_type))
    else:
        onboard_wallet_type = Account.WALLET_TYPE_FROM_STRING[onboard_wallet_type]

    if onboard_accept_eula != 'on':
        return render_template('xbr_onboard_submit_error.html', onboard_member_error='EULA must be accepted')

    # eg, onboard_wallet_address = 0x6231eECbA6e7983efe5ce6d16972E16cCcD97CE7
    if len(onboard_wallet_address) != 42:
        return render_template('xbr_onboard_submit_error.html',
                               onboard_member_error='Invalid wallet address "{}"'.format(onboard_wallet_address))

    try:
        onboard_wallet_address = binascii.a2b_hex(onboard_wallet_address[2:])
    except:
        return render_template('xbr_onboard_submit_error.html',
                               onboard_member_error='Invalid wallet address "{}"'.format(onboard_wallet_address))

    if not validate_email(onboard_member_email, check_mx=False, verify=False):
        return render_template('xbr_onboard_submit_error.html',
                               onboard_member_error='Invalid email address "{}"'.format(onboard_member_email))

    if not is_valid_username(onboard_member_name):
        return render_template(
            'xbr_onboard_submit_error.html',
            onboard_member_error='Invalid username "{}" - must be a string matching the regular expression {}'.format(
                onboard_member_name, _USERNAME_PAT_STR))

    db = app.config['DB']
    schema = app.config['DBSCHEMA']

    with db.begin() as txn:
        account_oid = schema.idx_accounts_by_username[txn, onboard_member_name]
        if account_oid:
            return render_template('xbr_onboard_submit_error.html',
                                   onboard_member_error='Username "{}" already exists'.format(onboard_member_name))

    vaction_oid = uuid.uuid4()
    vaction_code = generate_activation_code()
    mailgw = app.config['MAILGUN']
    try:
        mailgw.send_onboard_verification(onboard_member_email, vaction_oid, vaction_code)
    except Exception as e:
        return render_template('xbr_onboard_submit_error.html',
                               onboard_member_error='Failed to submit email via mailgun (exception {})'.format(e))

    on_success_url = '{}/member'.format(app.config['WEBSITE_URL'])
    on_error_url = None
    verified_data = {
        'onboard_member_name': onboard_member_name,
        'onboard_member_email': onboard_member_email,
        'onboard_wallet_type': onboard_wallet_type,
        'onboard_wallet_address': onboard_wallet_address,
        'on_success_url': on_success_url,
        'on_error_url': on_error_url,
    }

    with db.begin(write=True) as txn:
        # double check (again) for username collision, as the mailgun email submit happens async in above after
        # we initially checked for collision
        account_oid = schema.idx_accounts_by_username[txn, onboard_member_name]
        if account_oid:
            return render_template('xbr_onboard_submit_error.html',
                                   onboard_member_error='Username "{}" already exists'.format(onboard_member_name))

        vaction = VerifiedAction()
        vaction.oid = vaction_oid
        vaction.created = np.datetime64(time_ns(), 'ns')
        vaction.vtype = VerifiedAction.VERIFICATION_TYPE_ONBOARD_MEMBER
        vaction.vstatus = VerifiedAction.VERIFICATION_STATUS_PENDING
        vaction.vcode = vaction_code
        # vaction.verified_oid = None
        vaction.verified_data = verified_data

        schema.verified_actions[txn, vaction.oid] = vaction

    return render_template('xbr_onboard_submit_success.html',
                           onboard_member_email=onboard_member_email,
                           vaction_oid=vaction_oid)


# https://xbr.network/verify-action?code=82a996c7815e528a26bb21e81216c56df38a3eb06109134569ddc5cbdc77454a
@app.route('/verify-action/')
def page_xbr_verify_action():
    session['site_area'] = 'landing'
    session['site-page'] = None
    vaction_oid = request.args.get('action', None)
    vaction_code = request.args.get('code', None)
    member_onboarded = app.config['NETWORK'].verify_onboard_member(vaction_oid, vaction_code)
    print('member onboarded:', member_onboarded)
    return redirect('/member/')


@app.route('/member/<path:member_adr>')
def page_xbr_member(member_adr=None):
    session['site_area'] = 'landing'
    session['site-page'] = None
    member_adr = request.args.get('member_adr', None)
    return render_template('xbr_member.html', member_adr=member_adr)


@app.route('/market/<path:market_id>')
def page_xbr_market(market_id):
    session['site_area'] = 'landing'
    session['site-page'] = None
    return render_template('xbr_market.html', market_id=market_id)


@app.route('/market')
def page_xbr_market_query():
    market_id = request.args.get('market_id', '0x00000000000000000000000000000000')
    return page_xbr_market(market_id)


@app.route('/channel/<path:channel_adr>')
def page_xbr_channel(channel_adr):
    session['site_area'] = 'landing'
    session['site-page'] = None
    return render_template('xbr_channel.html', channel_adr=channel_adr)


@app.route('/channel')
def page_xbr_channel_query():
    channel_adr = request.args.get('channel_adr', '0x00000000000000000000000000000000')
    return page_xbr_channel(channel_adr)


@app.route('/address/<path:address>')
def page_xbr_address(address):
    session['site_area'] = 'landing'
    session['site-page'] = None
    return render_template('xbr_address.html', address=address)


@app.route('/address')
def page_xbr_address_query():
    address = request.args.get('address', '0x00000000000000000000000000000000')
    return page_xbr_address(address)


@app.route('/console')
def page_xbr_console():
    return render_template('xbr_console.html')


@app.route('/editor')
def page_editor():
    session['site_area'] = 'landing'
    session['site-page'] = None
    return render_template('test_editor.html')


if __name__ == "__main__":

    parser = argparse.ArgumentParser()

    parser.add_argument("-d", "--debug", action="store_true", help="Enable debug output.")
    parser.add_argument("-n",
                        "--nonetwork",
                        action="store_true",
                        help="No public internet/networking, disable external code (Google Analytics, Disqus).")
    parser.add_argument("-f", "--freeze", action="store_true", help="Enable freeze mode.")
    parser.add_argument("--port",
                        type=int,
                        default=8080,
                        help='Web port to use for embedded Web server. Use 0 to disable.')
    parser.add_argument("--mailgun_key",
                        type=str,
                        default=None,
                        help='Mailgun key (eg "key-00000000000000000000000000000000")')
    parser.add_argument("--mailgun_url",
                        type=str,
                        default=None,
                        help='Mailgun posting URL (eg "https://api.mailgun.net/v3/mailing.crossbar.io/messages")')
    options = parser.parse_args()

    if options.mailgun_key:
        app.config['MAILGUN_KEY'] = options.mailgun_key

    if options.mailgun_url:
        app.config['MAILGUN_URL'] = options.mailgun_url

    app.config['DEBUG'] = options.debug
    app.config['NONETWORK'] = options.nonetwork

    if options.freeze:
        # freeze mode
        from flask_frozen import Freezer
        freezer = Freezer(app)

        freezer.freeze()
    else:
        # dynamic serving mode
        log.startLogging(sys.stdout)
        resource = WSGIResource(reactor, reactor.getThreadPool(), app)
        site = Site(resource)
        site.noisy = False
        site.log = lambda _: None
        reactor.listenTCP(options.port, site)
        reactor.run()



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import json

from functools import partial

from twisted.internet.defer import inlineCallbacks
from twisted.web.http_headers import Headers

from autobahn.twisted.wamp import ApplicationSession
from autobahn.wamp.exception import ApplicationError
from autobahn.wamp.types import SubscribeOptions

from txaio import make_logger


class MessageForwarder(ApplicationSession):

    log = make_logger()

    def __init__(self, *args, **kwargs):
        self._webtransport = kwargs.pop("webTransport", None)

        if not self._webtransport:
            import treq
            self._webtransport = treq

        super(MessageForwarder, self).__init__(*args, **kwargs)

    @inlineCallbacks
    def onJoin(self, details):

        subscriptions = self.config.extra["subscriptions"]

        debug = self.config.extra.get("debug", False)
        method = self.config.extra.get("method", "POST")
        expectedCode = self.config.extra.get("expectedcode")

        @inlineCallbacks
        def on_event(url, *args, **kwargs):

            headers = Headers({b"Content-Type": [b"application/json"]})

            body = json.dumps({
                "args": args,
                "kwargs": kwargs
            },
                              sort_keys=False,
                              separators=(',', ':'),
                              ensure_ascii=False)

            # http://treq.readthedocs.org/en/latest/api.html#treq.request
            res = yield self._webtransport.request(method,
                                                   url.encode('utf8'),
                                                   data=body.encode('utf8'),
                                                   headers=headers)

            if expectedCode:
                if not res.code == expectedCode:
                    raise ApplicationError("Request returned {}, not the expected {}".format(res.code, expectedCode))

            if debug:
                content = yield self._webtransport.text_content(res)
                self.log.debug(content)

        for s in subscriptions:
            # Assert that there's "topic" and "url" entries
            assert "topic" in s
            assert "url" in s

            yield self.subscribe(partial(on_event, s["url"]),
                                 s["topic"],
                                 options=SubscribeOptions(match=s.get("match", "exact")))

            self.log.debug("MessageForwarder subscribed to {topic}", topic=s["topic"])



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from autobahn.twisted.wamp import ApplicationSession
from autobahn.wamp.types import PublishOptions

from twisted.internet.defer import inlineCallbacks

_ = []


class AppSession(ApplicationSession):
    @inlineCallbacks
    def onJoin(self, details):
        yield self.subscribe(_.append, "com.test")
        yield self.publish("com.test", "woo", options=PublishOptions(exclude_me=False))



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################
"""Crossbar.io is a decentralized data plane for XBR/WAMP based application
service and data routing, built on Crossbar.io OSS."""

import os
import sys

import psutil

from crossbar._util import hl
from crossbar._version import __version__, __build__

import warnings

# prevent error in EXE building (pyinstaller):
# "pkg_resources.DistributionNotFound: The 'humanize' distribution was not found and is required by the application"
import humanize  # noqa

# monkey patch, see:
# https://github.com/ethereum/web3.py/issues/1201
# https://github.com/ethereum/eth-abi/pull/88
from eth_abi import abi

if not hasattr(abi, 'collapse_type'):

    def collapse_type(base, sub, arrlist):
        return base + sub + ''.join(map(repr, arrlist))

    abi.collapse_type = collapse_type

if not hasattr(abi, 'process_type'):
    from eth_abi.grammar import (
        TupleType,
        normalize,
        parse,
    )

    def process_type(type_str):
        normalized_type_str = normalize(type_str)
        abi_type = parse(normalized_type_str)

        type_str_repr = repr(type_str)
        if type_str != normalized_type_str:
            type_str_repr = '{} (normalized to {})'.format(
                type_str_repr,
                repr(normalized_type_str),
            )

        if isinstance(abi_type, TupleType):
            raise ValueError("Cannot process type {}: tuple types not supported".format(type_str_repr, ))

        abi_type.validate()

        sub = abi_type.sub
        if isinstance(sub, tuple):
            sub = 'x'.join(map(str, sub))
        elif isinstance(sub, int):
            sub = str(sub)
        else:
            sub = ''

        arrlist = abi_type.arrlist
        if isinstance(arrlist, tuple):
            arrlist = list(map(list, arrlist))
        else:
            arrlist = []

        return abi_type.base, sub, arrlist

    abi.process_type = process_type

# https://stackoverflow.com/a/40846742/884770
# https://github.com/numpy/numpy/pull/432/commits/170ed4e33d6196d724dc18ddcd42311c291b4587?diff=split
# https://docs.python.org/3/library/warnings.html
# /opt/pypy3/lib-python/3/importlib/_bootstrap.py:223: builtins.UserWarning: builtins.type size changed, may indicate binary incompatibility. Expected 872, got 416
warnings.filterwarnings("ignore", message="builtins.type size changed, may indicate binary incompatibility")

# https://peps.python.org/pep-0632/
# site-packages/_distutils_hack/__init__.py:30: builtins.UserWarning: Setuptools is replacing distutils.
warnings.filterwarnings("ignore", message="Setuptools is replacing distutils.")

__all__ = ('__version__', '__build__', 'run', 'personalities')

_DEFINED_REACTORS = ['select', 'poll', 'epoll', 'kqueue', 'iocp']

_DEFINED_PERSONALITIES = ['standalone', 'edge', 'master']

_TOP_COMMANDS = ['standalone', 'edge', 'network', 'master', 'shell', 'quickstart', 'legal', 'version']

_HELP_USAGE = """
Usage: {executable} <command>

    <command>:

    standalone       Run a Crossbar.io node (default when no command is given)
    edge             Run a crossbar Edge node
    master           Run a crossbar Master node
    network          Run a crossbar XBR Network node
    shell            Run a management shell command
    quickstart       Create a WAMP/XBR application skeleton
    version          Print crossbar software versions
    legal            Print crossbar license terms

Command help: {executable} <command> --help
"""


def run():
    """
    CLI entry point into crossbar.

    * when called with no arguments, the arguments are read
      from the command line

    * when called with an explicit arguments list, use the same
      arguments in the list as on the command line, leaving out
      the initial `crossbar` command.


    twisted.internet.error.ReactorNotRestartable

    **Examples**

    To start Crossbar.io programmatically from a given node directory:

    .. code-block:: python

        import crossbar

        crossbar.run(['start', '--cbdir', '/tmp/mynode1/.crossbar'])

    To start Crossbar.io with log output at level "debug"

    .. code-block:: console

        $ crossbar start --loglevel=debug

    To chose a specific event reactor and print version information

    .. code-block:: console

        $ CROSSBAR_REACTOR="kqueue" crossbar version

    To start from a specified node directory (instead of the default `$CWD/.crossbar`):

    .. code-block:: console

        $ CROSSBAR_DIR="/tmp/test/.crossbar" crossbar start

    which is the same as

    .. code-block:: console

        $ crossbar start --cbdir=/tmp/test/.crossbar

   **Influential command line options and environment variables include:**

    ==================  ========================  ==============================================
    Command line arg:   Environment variable:     Description:
    ==================  ========================  ==============================================
    n.a.                **CROSSBAR_REACTOR**      Event reactor:

                                                  * **auto**
                                                  * **select**
                                                  * **poll**
                                                  * **epoll**
                                                  * **kqueue**
                                                  * **iocp**

    n.a.                **CROSSBAR_PERSONALITY**  Node personality:

                                                  * **standalone**
                                                  * **edge**
                                                  * **master**

    ``--cbdir``         **CROSSBAR_DIR**          Node directory (local directory)
    ``--config``        **CROSSBAR_CONFIG**       Node configuration (local filename)
    ``--color``         n.a.                      Enable colored terminal output
    ``--loglevel``      n.a.                      Select log level
    ``--logformat``     n.a.                      Select log format
    ``--logdir``        n.a.                      Log to this local directory
    ``--logtofile``     n.a.                      Enable logging to file
    ==================  ========================  ==============================================

    .. seealso:: `TwistedMatrix: Choosing a Reactor and GUI Toolkit Integration <https://twistedmatrix.com/documents/current/core/howto/choosing-reactor.html>`_

    .. note:: The Twisted reactor to use can only be explicitly chosen via an environment
        variable, not a command line option.
    """
    #
    # check command line args
    #
    argv = sys.argv[:]
    if len(argv) < 2:
        print(_HELP_USAGE.format(executable=hl(os.path.basename(argv[0]), bold=True, color='yellow')))
        sys.exit(0)

    # XXX maybe we should Click here, too, since we're already depending on it?
    executable, command = argv[0:2]
    args = argv[2:]
    sys.argv = [executable] + argv[2:]

    # if no known top-level command was given, fallback to "edge" mode
    if command not in _TOP_COMMANDS:
        args = [command] + args
        command = 'standalone'

    # redirect a plain "crossbar legal" to "crossbar master legal"
    if command == 'legal':
        command = 'master'
        args = ['legal']

    # redirect a plain "crossbar version" to "crossbar master version"
    if command == 'version':
        command = 'master'
        args = ['version']

    # shell command (using asyncio)
    if command == 'shell':

        from crossbar.shell import main

        sys.exit(main.run())

    elif command == 'quickstart':
        from crossbar.quickstart import main

        sys.exit(main.run())

    elif command == 'quickstart-venv':
        from crossbar.quickstart.quickstartvenv import main

        sys.exit(main.run())

    # edge/master commands (using Twisted)
    elif command in ['standalone', 'edge', 'master', 'network']:

        # FIXME :: having FX drop out due to lack of entropy makes things a lot
        #          harder from a deployment / monitoring perspective. A retry with reporting
        #          would make life a lot easier.

        # on Linux, check that we start with sufficient system entropy
        if sys.platform.startswith('linux'):
            try:
                with open('/proc/sys/kernel/random/entropy_avail', 'r') as ent:
                    entropy_avail = int(ent.read())
                    if entropy_avail < 64:
                        print(
                            'FATAL: cannot start due to insufficient entropy ({} bytes) available - try installing rng-tools'
                            .format(entropy_avail))
                        sys.exit(1)
            except PermissionError:
                # this happens when packaged as a snap: the code prevented from reading a location
                # # that is not allowed to a confined snap package
                entropy_avail = -1

        mem_avail = psutil.virtual_memory().available // 2**20
        if mem_avail < 100:
            print('FATAL: cannot start due to insufficient available memory ({} MB free)'.format(mem_avail))
            sys.exit(1)

        from autobahn.twisted import install_reactor

        # IMPORTANT: keep the reactor install as early as possible to avoid importing
        # any Twisted module that comes with the side effect of installing a default
        # reactor (which might not be what we want!).
        reactor = install_reactor(explicit_reactor=os.environ.get('CROSSBAR_REACTOR', None),
                                  verbose=False,
                                  require_optimal_reactor=False)

        # get chosen personality class
        if command == 'standalone':
            from crossbar import personality as standalone

            personality = standalone.Personality

        elif command == 'edge':

            from crossbar import edge

            personality = edge.Personality

        elif command == 'network':

            from crossbar import network

            personality = network.Personality

        elif command == 'master':
            from crossbar import master

            personality = master.Personality

        else:
            # should not arrive here
            raise Exception('internal error')

        # do NOT move this import above *** (triggers reactor imports)
        from crossbar.node.main import main

        # and now actually enter here .. this never returns!
        sys.exit(main(executable, args, reactor, personality))

    else:
        assert False, 'should not arrive here'


def personalities():
    """
    Return a map from personality names to actual available (=installed) Personality classes.
    """
    #
    # do NOT move the imports here to the module level! (triggers reactor imports)
    #
    from crossbar import personality as standalone

    personality_classes = {'standalone': standalone.Personality}

    try:
        from crossbar import edge  # noqa
    except ImportError:
        pass
    else:
        personality_classes['edge'] = edge.Personality

    try:
        from crossbar import master  # noqa
    except ImportError:
        pass
    else:
        personality_classes['master'] = master.Personality

    return personality_classes


if __name__ == '__main__':
    run()
